# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Handler for new product batches
"""

from __future__ import unicode_literals, absolute_import

from rattail.db import model
from rattail.batch import BatchHandler


class NewProductBatchHandler(BatchHandler):
    """
    Handler for new product batches.
    """
    batch_model_class = model.NewProductBatch

    def should_populate(self, batch):
        if batch.input_filename:
            return True
        return False

    def populate(self, batch, progress=None):
        if batch.input_filename:
            return self.populate_from_file(batch, progress=progress)

    def populate_from_file(self, batch, progress=None):
        raise NotImplementedError

    def refresh_row(self, row):
        model = self.model
        session = self.app.get_session(row)

        if not row.upc and row.item_entry:
            row.upc = self.app.make_gpc(row.item_entry)

        # vendor
        if not row.vendor:
            if row.vendor_id:
                row.vendor = session.query(model.Vendor)\
                                    .filter(model.Vendor.id == row.vendor_id)\
                                    .first()
        if row.vendor:
            row.vendor_id = row.vendor.id

        # department
        if not row.department:
            if row.department_number:
                row.department = session.query(model.Department)\
                                        .filter(model.Department.number == row.department_number)\
                                        .first()
            elif row.department_name:
                row.department = session.query(model.Department)\
                                        .filter(model.Department.name == row.department_name)\
                                        .first()
                if not row.department:
                    row.department = session.query(model.Department)\
                                            .filter(model.Department.name.ilike('%{}%'.format(row.department_name)))\
                                            .first()
        if row.department:
            row.department_number = row.department.number
            row.department_name = row.department.name

        # subdepartment
        if not row.subdepartment:
            if row.subdepartment_number:
                row.subdepartment = session.query(model.Subdepartment)\
                                           .filter(model.Subdepartment.number == row.subdepartment_number)\
                                           .first()
            elif row.subdepartment_name:
                row.subdepartment = session.query(model.Subdepartment)\
                                           .filter(model.Subdepartment.name == row.subdepartment_name)\
                                           .first()
                if not row.subdepartment:
                    row.subdepartment = session.query(model.Subdepartment)\
                                               .filter(model.Subdepartment.name.ilike('%{}%'.format(row.subdepartment_name)))\
                                               .first()
        if row.subdepartment:
            row.subdepartment_number = row.subdepartment.number
            row.subdepartment_name = row.subdepartment.name

        # category
        if not row.category:
            if row.category_code:
                row.category = session.query(model.Category)\
                                      .filter(model.Category.code == row.category_code)\
                                      .first()
        if row.category:
            row.category_code = row.category.code

        # family
        if not row.family:
            if row.family_code:
                row.family = session.query(model.Family)\
                                    .filter(model.Family.code == row.family_code)\
                                    .first()
        if row.family:
            row.family_code = row.family.code

        # report
        if not row.report:
            if row.report_code:
                row.report = session.query(model.ReportCode)\
                                    .filter(model.ReportCode.code == row.report_code)\
                                    .first()
        if row.report:
            row.report_code = row.report.code

        # brand
        if not row.brand:
            if row.brand_name:
                row.brand = session.query(model.Brand)\
                                   .filter(model.Brand.name == row.brand_name)\
                                   .first()
        if row.brand:
            row.brand_name = row.brand.name

        if not row.product:
            if not row.item_entry:
                row.status_code = row.STATUS_MISSING_KEY
                return

            row.product = self.locate_product_for_entry(
                session, row.item_entry,
                type2_lookup=row.batch.get_param('type2_lookup'))

        if row.product:
            row.status_code = row.STATUS_PRODUCT_EXISTS
            return

        if row.vendor_id and not row.vendor:
            row.status_code = row.STATUS_VENDOR_NOT_FOUND
            return

        if (row.department_number or row.department_name) and not row.department:
            row.status_code = row.STATUS_DEPT_NOT_FOUND
            return

        if ((row.subdepartment_number or row.subdepartment_name)
            and not row.subdepartment):
            row.status_code = row.STATUS_SUBDEPT_NOT_FOUND
            return

        if row.category_code and not row.category:
            row.status_code = row.STATUS_CATEGORY_NOT_FOUND
            return

        if row.family_code and not row.family:
            row.status_code = row.STATUS_FAMILY_NOT_FOUND
            return

        if row.report_code and not row.report:
            row.status_code = row.STATUS_REPORTCODE_NOT_FOUND
            return

        row.status_code = row.STATUS_OK



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Project Commands
"""

import os
import subprocess
import sys
from pathlib import Path

import typer
from typing_extensions import Annotated

from .base import rattail_typer, Subcommand
from .util import require_prompt_toolkit, require_rich, rprint, basic_prompt
from rattail.projects.handler import RattailProjectHandler


@rattail_typer.command()
def make_project(
        ctx: typer.Context,
        path: Annotated[
            Path,
            typer.Argument(help="Path to project folder")] = ...,
):
    """
    Make a new app project
    """
    config = ctx.parent.rattail_config
    do_make_project(config, path)


class MakeProject(Subcommand):
    """
    Make a new app project
    """
    name = 'make-project'
    description = __doc__.strip()

    def add_args(self):
        """ """
        super().add_args()
        parser = self.parser

        parser.add_argument('path',
                            help="Path to project folder")

    def run(self, args):
        do_make_project(self.config, args.path)


def do_make_project(config, path):

    path = os.path.abspath(path)
    if os.path.exists(path):
        sys.stderr.write("path already exists: {}\n".format(path))
        sys.exit(1)

    # nb. technically these would be required and auto-installed
    # as needed (later), but seems better to do this explicitly
    # up-front before any other command output
    require_prompt_toolkit()
    require_rich()

    # welcome, continue?
    rprint("\n\t[blue]Welcome to Rattail[/blue]")
    rprint("\n\tThis tool will generate source code for a new project.")
    if not basic_prompt("continue?", True, is_bool=True):
        rprint()
        sys.exit(0)

    # name
    name = os.path.basename(path)

    # app_table_prefix
    prefix = name
    prefix = prefix.replace('-', '_')
    prefix = prefix.replace(' ', '_')
    app_table_prefix = basic_prompt('app_table_prefix',
                                    default=prefix)
    app_table_prefix = app_table_prefix.rstrip('_')

    # app_class_prefix
    prefix = name
    prefix = prefix.replace('-', '_')
    prefix = prefix.replace('_', ' ')
    app_class_prefix = ''.join([w.capitalize() for w in prefix.split()])
    app_class_prefix = basic_prompt('app_class_prefix',
                                    default=app_class_prefix)

    # org_name
    org_name = basic_prompt('org_name', required=True)

    # pypi_name
    pypi_name = name
    pypi_name = pypi_name.replace('_', ' ')
    pypi_name = pypi_name.replace('-', ' ')
    pypi_name = '-'.join([w.capitalize()
                          for w in org_name.split() + pypi_name.split()])
    pypi_name = basic_prompt('pypi_name', default=pypi_name)

    # app_title
    app_title = name
    app_title = app_title.replace('-', ' ')
    app_title = app_title.replace('_', ' ')
    app_title = ' '.join([w.capitalize() for w in app_title.split()])

    # generate project
    project_handler = RattailProjectHandler(config)
    options = {
        'name': app_title,
        'slug': name,
        'organization': org_name,
        'python_project_name': pypi_name,
        'python_name': app_table_prefix,
        'app_class_prefix': app_class_prefix,
        'has_db': True,
        'extends_db': True,
        'has_web': True,

        # TODO: these should not be needed..?
        'has_web_api': False,
        'has_datasync': False,
        'integrates_catapult': False,
        'integrates_corepos': False,
        'integrates_locsms': False,
        'uses_fabric': False,
    }
    project_handler.generate_project('rattail', name, options, path=path)
    rprint(f"\n\tproject created at:  [bold green]{path}[/bold green]")

    # install pkg
    if basic_prompt("install project package?", is_bool=True, default=True):
        subprocess.check_call(['pip', 'install', '-e', path])
        rprint(f"\n\tpackage installed:  [bold green]{pypi_name}[/bold green]")

        rprint("\n\tinstall and configure the app with:")
        rprint(f"\n\t[blue]{name} -n install[/blue]")

    rprint()



# -*- coding: utf-8; -*-
"""add inventory total cost

Revision ID: b3432785a839
Revises: fd935205655d
Create Date: 2017-07-19 00:22:04.784889

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'b3432785a839'
down_revision = u'fd935205655d'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_inventory
    op.add_column('batch_inventory', sa.Column('total_cost', sa.Numeric(precision=9, scale=5), nullable=True))
    op.add_column('batch_inventory_row', sa.Column('total_cost', sa.Numeric(precision=9, scale=5), nullable=True))


def downgrade():

    # batch_inventory
    op.drop_column('batch_inventory_row', 'total_cost')
    op.drop_column('batch_inventory', 'total_cost')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Installer Commands
"""

import warnings

from rattail.commands import Subcommand
from rattail.mako import ResourceTemplateLookup


class InstallSubcommand(Subcommand):
    """
    Base class and default implementation for ``poser install``
    commands.  Note that there is no ``rattail install`` command.
    """
    name = 'install'
    install_handler_spec = 'rattail.install:InstallHandler'

    # nb. these must be explicitly set b/c config is not available
    # when running normally, e.g. `poser -n install`
    #app_title = "Poser"
    #app_package = 'poser'
    #app_eggname = 'Poser'
    #app_pypiname = 'Poser'

    def run(self, args):
        factory = self.app.load_object(self.install_handler_spec)
        handler = factory(self.config,
                          app_title=self.app_title,
                          app_package=self.app_package,
                          app_eggname=self.app_eggname,
                          app_pypiname=self.app_pypiname)
        handler.run()


class InstallerTemplateLookup(ResourceTemplateLookup):
    def __init__(self, *args, **kwargs):
        warnings.warn("InstallerTemplateLookup is deprecated; "
                      "please use ResourceTemplateLookup instead",
                      DeprecationWarning, stacklevel=2)
        super().__init__(*args, **kwargs)



# -*- coding: utf-8; -*-
"""add purchase_batch_row.invoice_total_calculated

Revision ID: 6118198dc4db
Revises: 03be1584f923
Create Date: 2019-03-07 13:45:36.687862

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '6118198dc4db'
down_revision = '03be1584f923'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # purchase_batch
    op.add_column('purchase_batch', sa.Column('invoice_total_calculated', sa.Numeric(precision=8, scale=2), nullable=True))

    # purchase_batch_row
    op.add_column('purchase_batch_row', sa.Column('invoice_total_calculated', sa.Numeric(precision=7, scale=2), nullable=True))


def downgrade():

    # purchase_batch_row
    op.drop_column('purchase_batch_row', 'invoice_total_calculated')

    # purchase_batch
    op.drop_column('purchase_batch', 'invoice_total_calculated')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models for Work Orders
"""

import sqlalchemy as sa
from sqlalchemy import orm

from rattail.db.model import Base, uuid_column
from rattail.time import make_utc


class WorkOrder(Base):
    """
    Represents a generic work order for a customer.
    """
    __tablename__ = 'workorder'
    __table_args__ = (
        sa.ForeignKeyConstraint(['customer_uuid'], ['customer.uuid'], 
                                name='workorder_fk_customer'),
    )

    uuid = uuid_column()

    id = sa.Column(sa.Integer(), nullable=False, doc="""
    Numeric ID for the work order.
    """)

    customer_uuid = sa.Column(sa.String(length=32), nullable=False)
    customer = orm.relationship(
        'Customer',
        doc="""
        Reference to the customer who requested the work.
        """,
        backref=orm.backref(
            'workorders',
            cascade_backrefs=False,
            doc="""
            Sequence of all work orders for this customer.
            """))

    estimated_total = sa.Column(sa.Numeric(precision=9, scale=2), nullable=True, doc="""
    Estimated total price to be charged to the customer, should the
    work order be fulfilled.
    """)

    date_submitted = sa.Column(sa.Date(), nullable=True, doc="""
    Date on which the work order was first "submitted" by the customer.
    """)

    date_received = sa.Column(sa.Date(), nullable=True, doc="""
    Date on which the org received the work order from the customer.
    """)

    date_released = sa.Column(sa.Date(), nullable=True, doc="""
    Date on which the org "released" (e.g. mailed) the work result
    back to the customer.
    """)

    date_delivered = sa.Column(sa.Date(), nullable=True, doc="""
    Date on which the work result was truly delivered back to the
    customer.
    """)

    notes = sa.Column(sa.Text(), nullable=True, doc="""
    Extra notes about the work order.
    """)

    status_code = sa.Column(sa.Integer(), nullable=False, doc="""
    Status code for the work order.
    """)

    status_text = sa.Column(sa.String(length=255), nullable=True, doc="""
    Text which may briefly explain the status code, if needed.
    """)

    def __str__(self):
        return "#{} for {}".format(self.id, self.customer)

    @property
    def id_str(self):
        if not self.id:
            return ''

        from rattail.batch import batch_id_str
        return batch_id_str(self.id)


class WorkOrderEvent(Base):
    """
    An event in the life of a work order
    """
    __tablename__ = 'workorder_event'
    __table_args__ = (
        sa.ForeignKeyConstraint(['workorder_uuid'], ['workorder.uuid'],
                                name='workorder_event_fk_workorder'),
        sa.ForeignKeyConstraint(['user_uuid'], ['user.uuid'],
                                name='workorder_event_fk_user'),
    )

    uuid = uuid_column()

    workorder_uuid = sa.Column(sa.String(length=32), nullable=False)
    workorder = orm.relationship(
        WorkOrder,
        doc="""
        Reference to the :class:`CustomerOrder` instance to which the item belongs.
        """,
        backref=orm.backref(
            'events',
            order_by='WorkOrderEvent.occurred',
            cascade='all, delete-orphan',
            cascade_backrefs=False))

    type_code = sa.Column(sa.Integer, nullable=False, doc="""
    Code specifying the type of event this is.
    """)

    occurred = sa.Column(sa.DateTime(), nullable=False, default=make_utc, doc="""
    Date and time when the event occurred.
    """)

    user_uuid = sa.Column(sa.String(length=32), nullable=False)
    user = orm.relationship(
        'User',
        doc="""
        User who was the "actor" for the event.
        """)

    note = sa.Column(sa.Text(), nullable=True, doc="""
    Optional note recorded for the event.
    """)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Console Commands
"""

import sys
import time
import datetime
import logging
from enum import Enum
from pathlib import Path

import typer
from typing_extensions import Annotated

from .base import rattail_typer, Subcommand


log = logging.getLogger(__name__)


class ServiceAction(str, Enum):
    start = 'start'
    stop = 'stop'
    check_queue = 'check_queue'
    check_watchers = 'check_watchers'
    collectd = 'collectd'
    wait = 'wait'
    remove_settings = 'remove_settings'


@rattail_typer.command()
def datasync(
        ctx: typer.Context,
        action: Annotated[
            ServiceAction,
            typer.Argument(help="Action to perform for the service.")] = ...,
        pidfile: Annotated[
            Path,
            typer.Option('--pidfile', '-p',
                         help="Path to PID file.")] = '/var/run/rattail/datasync.pid',
        daemonize: Annotated[
            bool,
            typer.Option(help="DEPRECATED")] = False,
        timeout: Annotated[
            int,
            typer.Option('--timeout', '-T',
                         help="Optional timeout (in minutes) for use with the 'wait' "
                         "or 'check' actions.  If specified for 'wait', the waiting "
                         "will stop after the given number of minutes and exit with a "
                         "nonzero code to indicate failure.  If specified for 'check', "
                         "the command will perform some health check based on the given "
                         "timeout, and exit with nonzero code if the check fails.")] = 0,
        collectd_plugin: Annotated[
            str,
            typer.Option(help="The `plugin-instance` part of the 'Identifier' "
                         "to be passed to collectd via PUTVAL entries.")] = 'datasync',
):
    """
    Manage the data sync daemon
    """
    from rattail.datasync.daemon import DataSyncDaemon

    config = ctx.parent.rattail_config

    if action == 'wait':
        do_wait(config, timeout)

    elif action == 'check':
        do_check_queue(config, timeout)

    elif action == 'check-queue':
        do_check_queue(config, timeout)

    elif action == 'check-watchers':
        do_check_watchers(config, timeout)

    elif action == 'collectd':
        do_collectd(config, collectd_plugin)

    elif action == 'remove-settings':
        do_remove_settings(config)

    else: # manage the daemon
        daemon = DataSyncDaemon(pidfile, config=config)
        if action == 'stop':
            daemon.stop()
        else: # start
            try:
                daemon.start(daemonize=False)
            except KeyboardInterrupt:
                sys.stderr.write("Interrupted.\n")


class DataSync(Subcommand):
    """
    Manage the data sync daemon
    """
    name = 'datasync'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        subparsers = parser.add_subparsers(title='subcommands')

        start = subparsers.add_parser('start', help="Start service")
        start.set_defaults(subcommand='start')

        stop = subparsers.add_parser('stop', help="Stop service")
        stop.set_defaults(subcommand='stop')

        check = subparsers.add_parser('check', help="(DEPRECATED) Check for stale (lingering) changes in queue")
        check.set_defaults(subcommand='check')

        check_queue = subparsers.add_parser('check-queue', help="Check for stale (lingering) changes in queue")
        check_queue.set_defaults(subcommand='check-queue')

        check_watchers = subparsers.add_parser('check-watchers', help="Check for dead watcher threads")
        check_watchers.set_defaults(subcommand='check-watchers')

        collectd = subparsers.add_parser('collectd', help="Output statistics for collectd")
        collectd.set_defaults(subcommand='collectd')

        wait = subparsers.add_parser('wait', help="Wait for changes to be processed")
        wait.set_defaults(subcommand='wait')

        remove_settings = subparsers.add_parser(
            'remove-settings',
            help="Remove all datasync settings from the DB")
        remove_settings.set_defaults(subcommand='remove-settings')

        parser.add_argument('-p', '--pidfile', metavar='PATH', default='/var/run/rattail/datasync.pid',
                            help="Path to PID file.")
        parser.add_argument('--daemonize', action='store_true',
                            help="DEPRECATED")
        parser.add_argument('--no-daemonize',
                            '-D', '--do-not-daemonize',
                            action='store_false', dest='daemonize',
                            help="DEPRECATED")
        parser.add_argument('-T', '--timeout', metavar='MINUTES', type=int, default=0,
                            help="Optional timeout (in minutes) for use with the 'wait' or 'check' commands.  "
                            "If specified for 'wait', the waiting still stop after the given number of minutes "
                            "and exit with a nonzero code to indicate failure.  If specified for 'check', the "
                            "command will perform some health check based on the given timeout, and exit with "
                            "nonzero code if the check fails.")

        parser.add_argument('--collectd-plugin', default='datasync',
                            help="The `plugin-instance` part of the 'Identifier' "
                            "to be passed to collectd via PUTVAL entries.")

    def run(self, args):
        from rattail.datasync.daemon import DataSyncDaemon

        self.datasync_handler = self.app.get_datasync_handler()

        if args.subcommand == 'wait':
            do_wait(self.config, args.timeout)

        elif args.subcommand == 'check':
            do_check_queue(self.config, args.timeout)

        elif args.subcommand == 'check-queue':
            do_check_queue(self.config, args.timeout)

        elif args.subcommand == 'check-watchers':
            do_check_watchers(self.config, args.timeout)

        elif args.subcommand == 'collectd':
            do_collectd(self.config, args.collectd_plugin)

        elif args.subcommand == 'remove-settings':
            do_remove_settings(self.config)

        else: # manage the daemon
            daemon = DataSyncDaemon(args.pidfile, config=self.config)
            if args.subcommand == 'stop':
                daemon.stop()
            else: # start
                try:
                    daemon.start(daemonize=False)
                except KeyboardInterrupt:
                    self.stderr.write("Interrupted.\n")


def do_check_queue(config, timeout):
    """
    Perform general queue / health check for datasync.
    """
    app = config.get_app()
    model = app.model
    session = app.make_session()

    # looking for changes which have been around for "timeout" minutes
    timeout = timeout or 90
    cutoff = app.make_utc() - datetime.timedelta(seconds=60 * timeout)
    changes = session.query(model.DataSyncChange)\
                     .filter(model.DataSyncChange.obtained < cutoff)\
                     .count()
    session.close()

    # if we found stale changes, then "fail" - otherwise we'll "succeed"
    if changes:
        # TODO: should log set of unique payload types, for troubleshooting
        log.debug("found %s changes, in queue for %s minutes", changes, timeout)
        sys.stderr.write("Found {} changes, in queue for {} minutes\n".format(changes, timeout))
        sys.exit(1)

    log.info("found no changes in queue for %s minutes", timeout)


def do_check_watchers(config, timeout):
    """
    Perform general health check for datasync watcher threads.
    """
    app = config.get_app()
    datasync = app.get_datasync_handler()
    profiles = datasync.get_configured_profiles()
    session = app.make_session()

    # cutoff is "timeout" minutes before "now"
    timeout = timeout or 15
    cutoff = app.make_utc() - datetime.timedelta(seconds=60 * timeout)

    dead = []
    for key in profiles:

        # looking for watcher "last run" time older than "timeout" minutes
        lastrun = datasync.get_watcher_lastrun(
            key, tzinfo=False, session=session)
        if lastrun and lastrun < cutoff:
            dead.append(key)

    session.close()

    # if we found dead watchers, then "fail" - otherwise we'll "succeed"
    if dead:
        sys.stderr.write("Found {} watcher threads dead for {} minutes: {}\n".format(len(dead), timeout, ', '.join(dead)))
        sys.exit(1)

    log.info("found no watcher threads dead for %s minutes", timeout)


def do_collectd(config, plugin):
    """
    Output statistics for collectd
    """
    app = config.get_app()
    datasync = app.get_datasync_handler()
    profiles = datasync.get_configured_profiles()
    session = app.make_session()
    model = app.model

    # current time (sanity check)
    now = app.localtime()
    collectd_putval(config, plugin, 'gauge-current_time',
                    now.timestamp(), now=now)

    # iterate watcher profiles
    for key in profiles:

        # last run for watcher
        lastrun = datasync.get_watcher_lastrun(
            key, local=True, tzinfo=True, session=session)
        collectd_putval(config, plugin, f'gauge-{key}_watcher_lastrun',
                        lastrun.timestamp(), now=now)

        # seconds since last run
        collectd_putval(config, plugin, f'gauge-{key}_watcher_seconds_since_lastrun',
                        now.timestamp() - lastrun.timestamp(), now=now)

    # now inspect the change queue..

    # total changes in queue
    total = session.query(model.DataSyncChange).count()
    collectd_putval(config, plugin, 'gauge-queue_changes_total',
                    total, now=now)

    # stale changes in queue
    timeout = config.getint('rattail.datasync',
                            'collectd.changes_stale_timeout',
                            default=15) # minutes
    cutoff = app.make_utc() - datetime.timedelta(seconds=60 * timeout)
    stale = session.query(model.DataSyncChange)\
                   .filter(model.DataSyncChange.obtained < cutoff)\
                   .count()
    collectd_putval(config, plugin, 'gauge-queue_changes_stale',
                    stale, now=now)
    collectd_putval(config, plugin, 'gauge-queue_changes_stale_timeout',
                    timeout, now=now)

    session.close()


def collectd_putval(config, plugin, data_type, value,
                    hostname=None, interval=None, now=None):
    app = config.get_app()
    if not hostname:
        hostname = app.get_collectd_hostname()

    if not interval:
        interval = app.get_collectd_interval()
    if interval:
        interval = ' interval={}'.format(interval)
    else:
        interval = ''

    if now:
        value = '{}:{}'.format(now.timestamp(), value)
    else:
        value = 'N:{}'.format(value)

    msg = 'PUTVAL {}/{}/{}{} {}\n'.format(
        hostname,
        plugin,
        data_type,
        interval,
        value)

    sys.stdout.write(msg)


def do_remove_settings(config):
    """
    Remove all datasync settings from the DB
    """
    from rattail.datasync.util import purge_datasync_settings

    app = config.get_app()
    session = app.make_session()
    purge_datasync_settings(config, session)
    session.commit()
    session.close()


def do_wait(config, timeout):
    app = config.get_app()
    model = app.model
    session = app.make_session()
    started = app.make_utc()
    log.debug("will wait for current change queue to clear")
    last_logged = started

    changes = session.query(model.DataSyncChange)
    count = changes.count()
    log.debug("there are %d changes in the queue", count)
    while count:
        try:
            now = app.make_utc()

            if timeout and (now - started).seconds >= (timeout * 60):
                log.warning("datasync wait timed out after %d minutes, with %d changes in queue",
                            timeout, count)
                sys.exit(1)

            if (now - last_logged).seconds >= 60:
                log.debug("still waiting, %d changes in the datasync queue", count)
                last_logged = now

            time.sleep(1)
            count = changes.count()

        except KeyboardInterrupt:
            sys.stderr.write("Waiting cancelled by user\n")
            session.close()
            sys.exit(1)

    session.close()
    log.debug("all datasync changes have been processed")



# -*- coding: utf-8; -*-
"""add person.preferred_first_name

Revision ID: 4e7b6d8e71a4
Revises: 11dd4ffbe8c9
Create Date: 2024-04-01 17:01:40.901202

"""

# revision identifiers, used by Alembic.
revision = '4e7b6d8e71a4'
down_revision = '11dd4ffbe8c9'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # person
    op.add_column('person', sa.Column('preferred_first_name', sa.String(length=50), nullable=True))
    op.add_column('person_version', sa.Column('preferred_first_name', sa.String(length=50), autoincrement=False, nullable=True))


def downgrade():

    # person
    op.drop_column('person_version', 'preferred_first_name')
    op.drop_column('person', 'preferred_first_name')



# -*- coding: utf-8; -*-

from unittest import TestCase

import sqlalchemy as sa

from rattail import clientele
from rattail.config import make_config


class TestClienteleHandler(TestCase):

    def setUp(self):
        self.config = make_config([], extend=False)
        self.app = self.config.get_app()
        self.handler = clientele.ClienteleHandler(self.config)
        self.engine = sa.create_engine('sqlite://')
        self.app.model.Base.metadata.create_all(bind=self.engine)

    def test_get_all_customers(self):
        model = self.app.model
        session = self.app.make_session(bind=self.engine)

        # add some customers
        session.add(model.Customer(name='Fred Flintstone'))
        session.add(model.Customer(name='Wilma Flintstone'))
        session.add(model.Customer(name='Barney Rubble', active_in_pos=False))

        # getting "all" will exclude inactive by default
        customers = self.handler.get_all_customers(session)
        self.assertEqual(len(customers), 2)

        # but we can request them too
        customers = self.handler.get_all_customers(session, include_inactive=True)
        self.assertEqual(len(customers), 3)

        session.commit()
        session.close()



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Report Commands
"""

import logging

from sqlalchemy import orm

from rattail.commands import Subcommand
from rattail.time import get_sunday, get_monday


log = logging.getLogger(__name__)


class ReportSubcommand(Subcommand):
    """
    Generate and email a report
    """
    report_key = None
    email_key = None
    simple_report_name = "(Unnamed)"

    def ensure_report_key(self):
        if not self.report_key:
            raise RuntimeError("must define {}.report_key".format(
                self.__class__.__name__))


class WeeklyReport(ReportSubcommand):
    """
    Generate and email a weekly report
    """

    def run(self, args):
        self.ensure_report_key()
        user = self.get_runas_user()
        run_weekly_report(self.config, self.report_key, self.simple_report_name,
                          email_key=self.email_key, user=user, progress=self.progress)


def run_weekly_report(config, report_key, simple_report_name, email_key=None, user=None,
                      progress=None):
    app = config.get_app()
    model = app.model
    session = app.make_session()

    # first must determine most recent complete Mon-Sun date range
    # TODO: should probably be more flexible about date range..
    today = app.today()
    sunday = get_sunday(today)
    monday = get_monday(sunday)
    start_date = monday
    end_date = sunday

    # determine naming for the report
    report_name = "{} {} thru {}".format(
        simple_report_name,
        start_date.strftime("%Y-%m-%d"),
        end_date.strftime("%Y-%m-%d"))

    try:
        # see if this report has already been ran
        output = session.query(model.ReportOutput)\
                        .filter(model.ReportOutput.report_type == report_key)\
                        .filter(model.ReportOutput.report_name == report_name)\
                        .one()

    except orm.exc.NoResultFound:

        # generate report file and commit result
        handler = app.get_report_handler()
        report = handler.get_report(report_key)
        params = {'start_date': start_date, 'end_date': end_date}
        if user:
            user = session.get(model.User, user.uuid)
        output = handler.generate_output(session, report, params, user,
                                         progress=progress)
        session.commit()

        # try to make url for report
        report_url = None
        base_url = config.base_url()
        if base_url:
            report_url = '{}/reports/generated/{}'.format(
                base_url, output.uuid)

        # send report output as email
        handler.email_output(report, output, email_key or report_key,
                             extra_data={'report_url': report_url})

    else:
        log.warning("report output already exists: %s", report_name)

    session.close()



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail (self) -> Rattail data import
"""

import warnings

from rattail import importing


class FromRattailSelfToRattail(importing.FromRattailHandler, importing.ToSQLAlchemyHandler):
    """
    Common base class for import handlers which read data from the Rattail DB
    for the sake of updating misc. other tables in that same DB.
    """
    host_key = 'self'
    local_key = 'rattail'

    @property
    def host_title(self):
        title = self.config.node_title(default="Rattail")
        return f"{title} (self)"

    @property
    def local_title(self):
        title = self.config.node_title(default="Rattail")
        return title

    def begin_local_transaction(self):
        self.session = self.host_session

        # ToRattailHandler would do this for us, but alas..we copy/pasted this
        if hasattr(self, 'runas_username') and self.runas_username:
            self.session.set_continuum_user(self.runas_username)

    def rollback_transaction(self):
        self.rollback_host_transaction()

    def commit_transaction(self):
        self.commit_host_transaction()


class FromRattailSelf(importing.FromSQLAlchemy):
    """
    Common base class for the "host" side of importers which read data from the
    Rattail DB for the sake of updating misc. other tables in that same DB.
    """


class FromRattailLocalToRattail(FromRattailSelfToRattail):
    """ DEPRECATED """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warnings.warn("FromRattailLocalToRattail is deprecated; "
                      "please use FromRattailSelfToRattail instead",
                      DeprecationWarning, stacklevel=2)


class FromRattailLocal(FromRattailSelf):
    """ DEPRECATED """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warnings.warn("FromRattailLocal is deprecated; "
                      "please use FromRattailSelf instead",
                      DeprecationWarning, stacklevel=2)



# -*- coding: utf-8; -*-
"""add batch_pos_row.void

Revision ID: dd1a440a93f1
Revises: 3ef6a60a8898
Create Date: 2023-09-30 19:54:34.473398

"""

# revision identifiers, used by Alembic.
revision = 'dd1a440a93f1'
down_revision = '3ef6a60a8898'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_pos_row
    # nb. hack to get nullable=False for void
    op.add_column('batch_pos_row', sa.Column('void', sa.Boolean(), nullable=True))
    row = sa.sql.table('batch_pos_row', sa.sql.column('void'))
    op.execute(row.update().values({'void': False}))
    op.alter_column('batch_pos_row', 'void', nullable=False, existing_type=sa.Boolean())


def downgrade():

    # batch_pos_row
    op.drop_column('batch_pos_row', 'void')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Application Configuration
"""

import importlib
import os
import re
import sys
import datetime
import configparser
import warnings
import logging
import logging.config

from wuttjamaican.conf import (WuttaConfig, WuttaConfigExtension,
                               make_config as wutta_make_config,
                               generic_default_files)
from wuttjamaican.util import (parse_bool as wutta_parse_bool,
                               parse_list as wutta_parse_list)

from rattail.util import load_entry_points, load_object
from rattail.exceptions import WindowsExtensionsNotInstalled, ConfigurationError
from rattail.files import temp_path


log = logging.getLogger(__name__)


def parse_bool(value):
    """
    Compatibility wrapper for
    :func:`wuttjamaican:wuttjamaican.util.parse_bool()`.

    This function will eventually be deprecated; new code should use
    the upstream function instead.
    """
    return wutta_parse_bool(value)


def parse_list(value):
    """
    Compatibility wrapper for
    :func:`wuttjamaican:wuttjamaican.util.parse_list()`.

    This function will eventually be deprecated; new code should use
    the upstream function instead.
    """
    return wutta_parse_list(value)


class RattailConfig(WuttaConfig):
    """
    Configuration for Rattail apps.

    A single instance of this class is created on app startup, by way
    of calling :func:`rattail.config.make_config()`.

    This class is based on
    :class:`~wuttjamaican:wuttjamaican.conf.WuttaConfig` but adds many
    methods specific to Rattail.

    Some of the customizations supplied by this class are described
    below.

    .. attribute:: versioning_has_been_enabled

       Flag indicating whether SQLAlchemy-Continuum versioning has
       been enabled for the running app.  This gets set when
       :func:`~rattail.db.config.configure_versioning()` happens.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('appname', 'rattail')
        defaults = kwargs.setdefault('defaults', {})
        defaults.setdefault('rattail.app.handler', 'rattail.app:AppHandler')
        super().__init__(*args, **kwargs)

        # this is false, unless/until it becomes true
        self.versioning_has_been_enabled = False

    @property
    def prioritized_files(self):
        """
        Backward-compatible property which just calls
        :meth:`~wuttjamaican:wuttjamaican.conf.WuttaConfig.get_prioritized_files()`.

        New code should use ``get_prioritized_files()`` instead of
        this property.
        """
        return self.get_prioritized_files()

    def setdefault(self, *args):
        """
        We override this method to support different calling signatures.

        :meth:`wuttjamaican:wuttjamaican.conf.WuttaConfig.setdefault()`
        normally expects just ``(key, value)`` args, but we also (for
        now) support the older style of ``(section, option, value)`` -
        *eventually* that will go away but probably not in the near
        future.
        """
        # figure out what sort of args were passed
        if len(args) == 2:
            key, value = args
        elif len(args) == 3:
            section, option, value = args
            key = f'{section}.{option}'
        else:
            raise ValueError("must pass either 2 args (key, value), "
                             "or 3 args (section, option, value)")

        # then do normal logic
        super().setdefault(key, value)

    def get(self, *args, **kwargs):
        """
        We override this method to support different calling signatures.

        :meth:`wuttjamaican:wuttjamaican.conf.WuttaConfig.get()`
        normally expects just ``(key, ...)`` args, but we also (for
        now) support the older style of ``(section, option, ...)`` -
        *eventually* that will go away but probably not in the near
        future.
        """
        # figure out what sort of args were passed
        if len(args) == 1:
            key = args[0]
        elif len(args) == 2:
            section, option = args
            key = f'{section}.{option}'
        else:
            raise ValueError("must pass either 1 arg (key), "
                             "or 2 args (section, option)")

        # then do normal logic
        return super().get(key, **kwargs)

    def getbool(self, *args, **kwargs):
        """
        Backward-compatible alias for
        :meth:`~wuttjamaican:wuttjamaican.conf.WuttaConfig.get_bool()`.

        New code should use ``get_bool()`` instead of this method.
        """
        # TODO: eventually
        # warnings.warn("config.getbool() method is deprecated; "
        #               "please use config.get_bool() instead",
        #               DeprecationWarning, stacklevel=2)
        return self.get_bool(*args, **kwargs)

    def getint(self, *args, **kwargs):
        """
        Backward-compatible alias for
        :meth:`~wuttjamaican:wuttjamaican.conf.WuttaConfig.get_int()`.

        New code should use ``get_int()`` instead of this method.
        """
        # TODO: eventually
        # warnings.warn("config.getint() method is deprecated; "
        #               "please use config.get_int() instead",
        #               DeprecationWarning, stacklevel=2)
        return self.get_int(*args, **kwargs)

    def getlist(self, *args, **kwargs):
        """
        Backward-compatible alias for
        :meth:`~wuttjamaican:wuttjamaican.conf.WuttaConfig.get_list()`.

        New code should use ``get_list()`` instead of this method.
        """
        # TODO: eventually
        # warnings.warn("config.getlist() method is deprecated; "
        #               "please use config.get_list() instead",
        #               DeprecationWarning, stacklevel=2)
        return self.get_list(*args, **kwargs)

    def parse_bool(self, value):
        """
        Convenience method around the
        :func:`~wuttjamaican:wuttjamaican.util.parse_bool()` function.

        Usage of this method is discouraged, at least until some more
        dust settles.  This probably belongs on the app handler
        instead so it can be overridden more easily.
        """
        # TODO: eventually
        # warnings.warn("config.parse_bool() method is deprecated; "
        #               "please use app.parse_bool() instead",
        #               DeprecationWarning, stacklevel=2)
        return wutta_parse_bool(value)

    def parse_list(self, value):
        """
        Convenience method around the
        :func:`~wuttjamaican:wuttjamaican.util.parse_list()` function.

        Usage of this method is discouraged, at least until some more
        dust settles.  This probably belongs on the app handler
        instead so it can be overridden more easily.
        """
        # TODO: eventually
        # warnings.warn("config.parse_list() method is deprecated; "
        #               "please use app.parse_list() instead",
        #               DeprecationWarning, stacklevel=2)
        return wutta_parse_list(value)

    def make_list_string(self, values):
        """
        Coerce the given list of values to a string, for config
        storage.  If this string is later parsed via
        :meth:`parse_list()` then it should return the same list of
        values.

        For example::

           string = config.make_list_string(['foo', 'bar'])

           assert string == 'foo, bar'

           values = config.parse_list(string)

           assert values == ['foo', 'bar']
        """
        final = []
        for value in values:
            if ' ' in value:
                quote = '"' if "'" in value else "'"
                value = f"{quote}{value}{quote}"
            final.append(value)
        return ', '.join(final)

    def beaker_invalidate_setting(self, name):
        """
        Backward-compatible method for unused Beaker caching logic.

        This method has no effect and should not be used.
        """
        # TODO: eventually
        # warnings.warn("config.beaker_invalidate_setting() method is deprecated",
        #               DeprecationWarning, stacklevel=2)

    def node_type(self, default=None):
        """
        Returns the "type" of current node.  What this means will
        generally depend on the app logic.
        """
        try:
            return self.require('rattail', 'node_type', usedb=False)
        except ConfigurationError:
            if default:
                return default
            raise

    def production(self):
        """
        Returns boolean indicating whether the app is running in
        production mode
        """
        return self.getbool('rattail', 'production', default=False)

    def get_model(self):
        """
        Returns a reference to configured 'model' module; defaults to
        :mod:`rattail.db.model`.
        """
        spec = self.get('rattail', 'model', usedb=False,
                        default='rattail.db.model')
        return importlib.import_module(spec)

    def get_enum(self, **kwargs):
        """
        Returns a reference to configured "enum" module; defaults to
        :mod:`rattail.enum`.
        """
        kwargs.setdefault('usedb', False)
        spec = self.get('rattail', 'enum', default='rattail.enum', **kwargs)
        return importlib.import_module(spec)

    def get_trainwreck_model(self):
        """
        Returns a reference to the configured data 'model' module for
        Trainwreck.  Note that there is *not* a default value for
        this; it must be configured.
        """
        spec = self.require('rattail.trainwreck', 'model', usedb=False)
        return importlib.import_module(spec)

    def versioning_enabled(self):
        """
        Returns boolean indicating whether data versioning is enabled.
        """
        return self.getbool('rattail.db', 'versioning.enabled', usedb=False,
                            default=False)

    def getdate(self, *args, **kwargs):
        """
        Retrieve a date value from config.
        """
        value = self.get(*args, **kwargs)
        app = self.get_app()
        return app.parse_date(value)

    def product_key(self, **kwargs):
        """
        Deprecated; instead please see
        :meth:`rattail.app.AppHandler.get_product_key_field()`.
        """
        warnings.warn("config.product_key() is deprecated; please "
                      "use app.get_product_key_field() instead",
                      DeprecationWarning, stacklevel=2)
        return self.get_app().get_product_key_field()

    def product_key_title(self, key=None):
        """
        Deprecated; instead please see
        :meth:`rattail.app.AppHandler.get_product_key_label()`.
        """
        warnings.warn("config.product_key_title() is deprecated; please "
                      "use app.get_product_key_label() instead",
                      DeprecationWarning, stacklevel=2)
        return self.get_app().get_product_key_label(field=key)

    def app_package(self, default=None):
        """
        Returns the name of Python package for the top-level app.
        """
        if not default:
            return self.require('rattail', 'app_package')
        return self.get('rattail', 'app_package', default=default)

    def app_title(self, **kwargs):
        """ DEPRECATED """
        # TODO: should put a deprecation warning here, but it could
        # make things noisy for a while and i'm not ready for that
        app = self.get_app()
        return app.get_title(**kwargs)

    def node_title(self, **kwargs):
        """ DEPRECATED """
        # TODO: should put a deprecation warning here, but it could
        # make things noisy for a while and i'm not ready for that
        app = self.get_app()
        return app.get_node_title(**kwargs)

    def running_from_source(self):
        """
        Returns boolean indicating whether the app is running from
        source, as opposed to official release.
        """
        return self.getbool('rattail', 'running_from_source', default=False)

    def demo(self):
        """
        Returns boolean indicating whether the app is running in demo mode
        """
        return self.getbool('rattail', 'demo', default=False)

    def appdir(self, require=True, **kwargs):
        """
        Returns path to the 'app' dir, if known.
        """
        if require:
            path = os.path.join(sys.prefix, 'app')
            kwargs.setdefault('default', path)
        kwargs.setdefault('usedb', False)
        return self.get('rattail', 'appdir', **kwargs)

    def datadir(self, require=True):
        """
        Returns path to the 'data' dir, if known.
        """
        get = self.require if require else self.get
        return get('rattail', 'datadir')

    def workdir(self, require=True):
        """
        Returns path to the 'work' dir, if known.
        """
        get = self.require if require else self.get
        return get('rattail', 'workdir')

    def batch_filedir(self, key=None):
        """
        Returns path to root folder where batches (optionally of type
        'key') are stored.
        """
        path = os.path.abspath(self.require('rattail', 'batch.files'))
        if key:
            return os.path.join(path, key)
        return path

    def batch_filepath(self, key, uuid, filename=None, makedirs=False):
        """
        Returns absolute path to a batch's data folder, with optional
        filename appended.  If ``makedirs`` is set, the batch data
        folder will be created if it does not already exist.
        """
        rootdir = self.batch_filedir(key)
        filedir = os.path.join(rootdir, uuid[:2], uuid[2:])
        if makedirs and not os.path.exists(filedir):
            os.makedirs(filedir)
        if filename:
            return os.path.join(filedir, filename)
        return filedir

    def export_filedir(self, key=None):
        """
        Returns path to root folder where exports (optionally of type
        'key') are stored.
        """
        path = self.get('rattail', 'export.files')
        if not path:
            path = os.path.join(self.appdir(), 'data', 'exports')
        path = os.path.abspath(path)
        if key:
            return os.path.join(path, key)
        return path

    def export_filepath(self, key, uuid, filename=None, makedirs=False):
        """
        Returns absolute path to export data file, generated from the given args.
        """
        rootdir = self.export_filedir(key)
        filedir = os.path.join(rootdir, uuid[:2], uuid[2:])
        if makedirs and not os.path.exists(filedir):
            os.makedirs(filedir)
        if filename:
            return os.path.join(filedir, filename)
        return filedir

    def upgrade_filedir(self):
        """
        Returns path to root folder where upgrade files are stored.
        """
        path = os.path.abspath(self.require('rattail.upgrades', 'files'))
        return path

    def upgrade_filepath(self, uuid, filename=None, makedirs=False):
        """
        Returns absolute path to upgrade data file, generated from the given args.
        """
        rootdir = self.upgrade_filedir()
        filedir = os.path.join(rootdir, uuid[:2], uuid[2:])
        if makedirs and not os.path.exists(filedir):
            os.makedirs(filedir)
        if filename:
            return os.path.join(filedir, filename)
        return filedir

    def upgrade_command(self, default='/bin/sleep 30'):
        """
        Returns command to be used when performing upgrades.
        """
        # TODO: what were those reasons then..?
        # NOTE: we don't allow command to be specified in DB, for
        # security reasons..
        return self.getlist('rattail.upgrades', 'command', usedb=False,
                            default=default)

    def base_url(self):
        """
        Returns the configured "base" (root) URL for the web app.
        """
        # first try "generic" config option
        url = self.get('rattail', 'base_url')

        # or use tailbone as fallback, since it's most likely
        if url is None:
            url = self.get('tailbone', 'url.base')
            if not url:
                url = self.get('tailbone', 'url', ignore_ambiguous=True)
                if url:
                    warnings.warn(f"URGENT: instead of 'tailbone.url', "
                                  f"you should set 'tailbone.url.base'",
                                  DeprecationWarning, stacklevel=2)

        if url is not None:
            return url.rstrip('/')

    def datasync_url(self, **kwargs):
        """
        Returns configured URL for managing datasync daemon.
        """
        return self.get('rattail.datasync', 'url', **kwargs)

    def single_store(self):
        """
        Returns boolean indicating whether the system is configured to behave
        as if it belongs to a single Store.
        """
        return self.getbool('rattail', 'single_store', default=False)

    def get_store(self, session):
        """
        Returns a :class:`rattail.db.model.Store` instance
        corresponding to app config, or ``None``.
        """
        store = self.get('rattail', 'store')
        if store:
            app = self.get_app()
            org_handler = app.get_org_handler()
            return org_handler.get_store(session, store)


class ConfigExtension(WuttaConfigExtension):
    """
    Base class for all config extensions.

    This is just a compatibility wrapper around
    :class:`wuttjamaican:wuttjamaican.conf.WuttaConfigExtension`; new
    code should probably use that directly.
    """


def rattail_default_files(appname):
    """
    This is used in place of upstream
    :func:`wuttjamaican:wuttjamaican.conf.generic_default_files()` to
    customize the default files when none are specified at startup.

    Rattail has traditionally used
    e.g. ``/path/to/venv/app/quiet.conf`` as its "preferred default
    file" when running ad-hoc commands.  So this function will look
    for that file and return it if found; otherwise it just calls the
    upstream function.
    """
    # try to guess a default config path
    # TODO: for now, prefer app/quiet.conf if present, but
    # probably we should look for adhoc.conf instead, since
    # the point of this magic is to make running ad-hoc
    # commands easier..
    quiet = os.path.join(sys.prefix, 'app', 'quiet.conf')
    if os.path.exists(quiet):
        # this config is definitely app-specific
        return [quiet]

    return generic_default_files(appname)


def make_config(
        files=None,
        plus_files=None,
        versioning=None,
        **kwargs):
    """
    Make a new config object (presumably for global use), initialized
    per the given parameters and (usually) further modified by all
    registered config extensions.

    This is a wrapper around upstream
    :func:`wuttjamaican:wuttjamaican.conf.make_config()`; see those
    docs for most of the param descriptions.  Rattail customizes the
    logic as follows:

    :param versioning: Controls whether or not the versioning system
       is configured with the new config object.  If ``True``,
       versioning will be configured.  If ``False`` then it will not
       be configured.  If ``None`` (the default) then versioning will
       be configured only if the config values say that it should be.

    :returns: An instance of :class:`RattailConfig`.
    """
    # turn on display of rattail deprecation warnings by default
    # TODO: this should be configurable, and possibly live elsewhere?
    warnings.filterwarnings('default', category=DeprecationWarning,
                            module=r'^rattail')
    warnings.filterwarnings('default', category=DeprecationWarning,
                            module=r'^tailbone')
    warnings.filterwarnings('default', category=DeprecationWarning,
                            module=r'^wutt')

    # prep kwargs
    kwargs.setdefault('appname', 'rattail')
    kwargs.setdefault('default_files', rattail_default_files)
    kwargs.setdefault('factory', RattailConfig)

    # remove deprecated args
    kwargs.pop('use_wuttaconfig', None)

    # make config object
    config = wutta_make_config(files=files, plus_files=plus_files, **kwargs)
    log.debug("config files were: %s", config.files_read)

    if config.getbool('rattail', 'suppress_psycopg2_wheel_warning', usedb=False):
        # TODO: revisit this, does it require action from us?
        # suppress this warning about psycopg2 wheel; not sure what it means yet
        # exactly but it's causing frequent noise for us...
        warnings.filterwarnings(
            'ignore',
            r'^The psycopg2 wheel package will be renamed from release 2\.8; in order to keep '
            r'installing from binary please use "pip install psycopg2-binary" instead\. For details '
            r'see: <http://initd.org/psycopg/docs/install.html#binary-install-from-pypi>\.',
            UserWarning,
            r'^psycopg2$',
        )

    # maybe configure versioning
    if versioning is None:
        versioning = config.versioning_enabled()
    if versioning:
        from rattail.db.config import configure_versioning
        configure_versioning(config)

    # maybe set "future" behavior for SQLAlchemy
    if config.getbool('rattail.db', 'sqlalchemy_future_mode', usedb=False):
        from rattail.db import Session
        if Session:
            Session.configure(future=True)

    return config


def get_user_dir(create=False):
    """
    Returns a path to the "preferred" user-level folder, in which additional
    config files (etc.) may be placed as needed.  This essentially returns a
    platform-specific variation of ``~/.rattail/``.

    If ``create`` is ``True``, then the folder will be created if it does not
    already exist.
    """
    if sys.platform == 'win32':

        # Use the Windows Extensions libraries to fetch official defaults.
        try:
            from win32com.shell import shell, shellcon
        except ImportError:
            raise WindowsExtensionsNotInstalled
        else:
            path = os.path.join(shell.SHGetSpecialFolderPath(
                0, shellcon.CSIDL_APPDATA), 'rattail')

    else:
        path = os.path.expanduser('~/.rattail')

    if create and not os.path.exists(path):
        os.mkdir(path)
    return path


def get_user_file(filename, createdir=False):
    """
    Returns a full path to a user-level config file location.  This is obtained
    by first calling :func:`get_user_dir()` and then joining the result with
    ``filename``.

    The ``createdir`` argument will be passed to :func:`get_user_dir()` as its
    ``create`` arg, and may be used to ensure the user-level folder exists.
    """
    return os.path.join(get_user_dir(create=createdir), filename)


class ConfigProfile(object):
    """
    Generic class to represent a config "profile", as used by the filemon and
    datasync daemons, etc.

    .. todo::

       This clearly needs more documentation.

    .. attribute:: config

       Reference to the primary Rattail config object for the running app.

    .. attribute:: key

       String identifier unique to this profile, within the broader
       config section.
    """

    def __init__(self, config, key, **kwargs):
        self.config = config
        self.app = self.config.get_app()
        self.model = self.config.get_model()
        self.enum = self.config.get_enum()
        self.key = key
        self.prefix = kwargs.pop('prefix', key)
        self.load()

    def load(self):
        """
        Read all relevant settings etc. from the config object,
        setting attributes on this profile instance as needed.
        """

    def load_defaults(self):
        """
        Read all "default" (common) settings from config, for the
        current profile.
        """
        self.workdir = self._config_string('workdir')
        self.stop_on_error = self._config_boolean('stop_on_error', False)

    def load_actions(self):
        """
        Read the "actions" from config, for the current profile, and
        assign the result to ``self.actions``.
        """
        self.actions = []
        for action in self._config_list('actions'):
            self.actions.append(self._config_action(action))

    @property
    def section(self):
        """
        Each subclass of ``ConfigProfile`` must define this.
        """
        raise NotImplementedError

    def _config_string(self, option, **kwargs):
        return self.config.get(self.section,
                               '{}.{}'.format(self.prefix, option),
                               **kwargs)

    def _config_boolean(self, option, default=None):
        return self.config.getbool(self.section,
                                   '{}.{}'.format(self.prefix, option),
                                   default=default)

    def _config_int(self, option, minimum=1, default=None):
        """
        Retrieve the *integer* value for the given option.
        """
        option = '{}.{}'.format(self.prefix, option)

        # try to read value from config
        value = self.config.getint(self.section, option)
        if value is not None:

            # found a value; validate it
            if value < minimum:
                log.warning("config value %s is too small; falling back to minimum "
                            "of %s for option: %s", value, minimum, option)
                value = minimum

        # or, use default value, if valid
        elif default is not None and default >= minimum:
            value = default

        # or, just use minimum value
        else:
            value = minimum

        return value

    def _config_list(self, option, default=None, **kwargs):
        value = self._config_string(option, **kwargs)
        if value:
            return self.config.parse_list(value)

        if isinstance(default, list):
            return default

        return []

    def _config_action(self, name):
        """
        Retrieve an "action" value from config, for the current
        profile.  This returns a :class:`ConfigProfileAction`
        instance.
        """
        from rattail.monitoring import CommandAction

        function = self._config_string('action.{}.func'.format(name))
        class_ = self._config_string('action.{}.class'.format(name))
        cmd = self._config_string('action.{}.cmd'.format(name))

        specs = [1 if spec else 0 for spec in (function, class_, cmd)]
        if sum(specs) != 1:
            raise ConfigurationError(
                "Monitor profile '{}' (action '{}') must have exactly one of: "
                "function, class, command".format(self.prefix, name))

        action = ConfigProfileAction()
        action.config = self.config

        if function:
            action.spec = function
            action.action = load_object(action.spec)
        elif class_:
            action.spec = class_
            action.action = load_object(action.spec)(self.config)
        elif cmd:
            action.spec = cmd
            action.action = CommandAction(self.config, cmd)

        action.args = self._config_list('action.{}.args'.format(name))

        action.kwargs = {}
        pattern = re.compile(r'^{}\.action\.{}\.kwarg\.(?P<keyword>\w+)$'.format(self.prefix, name), re.IGNORECASE)
        settings = self.config.get_dict(self.section)
        for key in settings:
            match = pattern.match(key)
            if match:
                action.kwargs[match.group('keyword')] = settings[key]

        action.retry_attempts = self._config_int('action.{}.retry_attempts'.format(name), minimum=1)
        action.retry_delay = self._config_int('action.{}.retry_delay'.format(name), minimum=0)
        return action


class ConfigProfileAction(object):
    """
    Simple class to hold configuration for a particular "action"
    defined within a monitor :class:`ConfigProfile`.  Each instance
    has the following attributes:

    .. attribute:: spec

       The original "spec" string used to obtain the action callable.

    .. attribute:: action

       A reference to the action callable.

    .. attribute:: args

       A sequence of positional arguments to be passed to the callable
       (in addition to the file path) when invoking the action.

    .. attribute:: kwargs

       A dictionary of keyword arguments to be passed to the callable
       (in addition to the positional arguments) when invoking the
       action.

    .. attribute:: retry_attempts

       Number of attempts to make when invoking the action.  Defaults
       to ``1``, meaning the first attempt will be made but no retries
       will happen.

    .. attribute:: retry_delay

       Number of seconds to pause between retry attempts, if
       :attr:`retry_attempts` is greater than one.  Defaults to ``0``.
    """
    spec = None
    action = None
    args = []
    kwargs = {}
    retry_attempts = 1
    retry_delay = 0


class FreeTDSLoggingFilter(logging.Filter):
    """
    Custom logging filter, to suppress certain "write to server failed"
    messages relating to FreeTDS database connections.  They seem harmless and
    just cause unwanted error emails.
    """

    def __init__(self, *args, **kwargs):
        logging.Filter.__init__(self, *args, **kwargs)
        self.pattern = re.compile(r'(?:Read from|Write to) the server failed')

    def filter(self, record):
        if (record.name == 'sqlalchemy.pool.QueuePool'
            and record.funcName == '_finalize_fairy'
            and record.levelno == logging.ERROR
            and record.msg == "Exception during reset or similar"
            and record.exc_info
            and self.pattern.search(str(record.exc_info[1]))):

            # Log this as a warning instead of error, to cut down on our noise.
            record.levelno = logging.WARNING
            record.levelname = 'WARNING'

        return True



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Problem Reports for Rattail Systems
"""

import datetime

from sqlalchemy import orm

from rattail.problems import ProblemReport


class RattailProblemReport(ProblemReport):
    """
    Base class for problem reports pertaining to a Rattail systems.
    """
    system_key = 'rattail'


class PendingProducts(RattailProblemReport):
    """
    Shows list of pending product records, if any present.
    """
    problem_key = 'pending_products'
    problem_title = "Pending products"

    def find_problems(self, **kwargs):
        session = self.app.make_session()
        model = self.model

        pending = session.query(model.PendingProduct)\
                         .filter(model.PendingProduct.status_code == self.enum.PENDING_PRODUCT_STATUS_PENDING)\
                         .all()

        ready = session.query(model.PendingProduct)\
                       .filter(model.PendingProduct.status_code == self.enum.PENDING_PRODUCT_STATUS_READY)\
                       .all()

        session.close()
        problems = pending + ready
        return problems

    def get_email_context(self, problems, **kwargs):
        kwargs = super().get_email_context(problems, **kwargs)

        kwargs['products_handler'] = self.app.get_products_handler()

        url = self.config.base_url()
        if url:
            url = f'{url}/products/pending/'
        kwargs['url'] = url

        return kwargs


class ProductWithoutPrice(RattailProblemReport):
    """
    Looks for products which have null (or $0) regular price.
    """
    problem_key = 'product_without_price'
    problem_title = "Products with no price"

    def find_problems(self, **kwargs):
        problems = []
        session = self.app.make_session()
        model = self.model
        products = session.query(model.Product)\
                          .order_by(model.Product.upc)\
                          .options(orm.joinedload(model.Product.brand))\
                          .options(orm.joinedload(model.Product.department))\
                          .options(orm.joinedload(model.Product.regular_price))

        def inspect(product, i):
            price = product.regular_price
            if not price or not price.price:
                problems.append(product)

        self.progress_loop(inspect, products,
                           message="Looking for products with no price")
        session.close()
        return problems


class StaleInventoryBatch(RattailProblemReport):
    """
    Looks for "stale" inventory batches, those which were created but
    not executed within a certain amount of time.
    """
    problem_key = 'stale_inventory_batch'
    problem_title = "Stale inventory batches"

    def __init__(self, *args, **kwargs):
        super(StaleInventoryBatch, self).__init__(*args, **kwargs)

        self.cutoff_days = self.config.getint(
            'rattail', 'problems.stale_inventory_batches.cutoff_days',
            default=4)

    def find_problems(self, **kwargs):
        session = self.app.make_session()
        model = self.model

        today = self.app.localtime().date()
        cutoff = today - datetime.timedelta(days=self.cutoff_days)
        cutoff = datetime.datetime.combine(cutoff, datetime.time(0))
        cutoff = self.app.localtime(cutoff)

        batches = session.query(model.InventoryBatch)\
                         .filter(model.InventoryBatch.executed == None)\
                         .filter(model.InventoryBatch.created <= self.app.make_utc(cutoff))\
                         .options(orm.joinedload(model.InventoryBatch.created_by)\
                                  .joinedload(model.User.person))\
                         .all()

        session.close()
        return batches

    def get_email_context(self, problems, **kwargs):
        kwargs = super(StaleInventoryBatch, self).get_email_context(problems,
                                                                    **kwargs)
        kwargs['cutoff_days'] = self.cutoff_days
        return kwargs


class UpgradePending(RattailProblemReport):
    """
    Looks for any system upgrades which have yet to be executed.
    """
    problem_key = 'upgrade_pending'
    problem_title = "Pending upgrade"

    def find_problems(self, **kwargs):
        session = self.app.make_session()
        model = self.model
        upgrades = session.query(model.Upgrade)\
                          .filter(model.Upgrade.status_code == self.enum.UPGRADE_STATUS_PENDING)\
                          .options(orm.joinedload(model.Upgrade.created_by)\
                                   .joinedload(model.User.person))\
                          .all()
        session.close()
        return upgrades



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Membership Handler
"""

import warnings

from rattail.app import GenericHandler


class MembershipHandler(GenericHandler):
    """
    Base class and default implementation for membership handlers.
    """

    def max_one_per_person(self):
        """
        Check whether a person is allowed to have just one member
        account, or if multiple are allowed.

        :returns: Boolean; if true then only one member account is
           allowed per person; if false then multiple are allowed.
        """
        return self.config.getbool('rattail', 'members.max_one_per_person',
                                   default=False)

    def ensure_member(self, person, **kwargs):
        """
        Returns the Member record associated with the given person, creating
        it first if necessary.
        """
        member = self.get_member(person)
        if member:
            return member

        return self.make_member(person, **kwargs)

    def make_member(self, person, **kwargs):
        """
        Make and return a new Member instance.
        """
        raise NotImplementedError

    def begin_membership(self, member, **kwargs):
        """
        Begin an active membership.
        """
        raise NotImplementedError

    def get_member(self, obj):
        """
        Returns the member associated with the given person, if there is one.
        """
        model = self.model

        if isinstance(obj, model.Member):
            return obj

        elif isinstance(obj, model.Customer):
            if obj.members:
                return obj.members[0]

        else:
            person = self.app.get_person(obj)
            if person and person.members:
                return person.members[0]

    def get_members_for_account_holder(
            self,
            person,
            **kwargs
    ):
        """
        Return all Member records for which the given Person is the
        account holder.
        """
        return list(person.members)

    def get_customer(self, obj):
        """
        Returns the customer associated with the given member, if there is one.
        """
        warnings.warn("MembershipHandler.get_customer() is deprecated; "
                      "please use AppHandler.get_customer() instead")

        return self.app.get_customer(obj)

    def get_person(self, obj):
        """
        Returns the person associated with the given member, if there is one.
        """
        warnings.warn("MembershipHandler.get_person() is deprecated; "
                      "please use AppHandler.get_person() instead")

        return self.app.get_person(obj)

    def get_last_patronage_date(self, member, **kwargs):
        raise NotImplementedError

    def get_equity_full_investment_amount(self, **kwargs):
        """
        Should return the amount required for an account to become
        fully invested.
        """
        raise NotImplementedError

    def get_equity_total(self, member, cached=True, **kwargs):
        """
        Get the official equity total for the given member account.
        """
        if cached:
            return member.equity_total

        return sum([payment.amount for payment in member.equity_payments])



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Utilities
"""

import collections
import importlib
import re
import shlex
import unicodedata
import datetime
import decimal
import subprocess
import warnings

from wuttjamaican.util import (UNSPECIFIED,
                               load_entry_points as wutta_load_entry_points,
                               load_object as wutta_load_object)


# TODO: deprecate / remove this
NOTSET = UNSPECIFIED


class OrderedDict(collections.OrderedDict):
    def __init__(self, *args, **kwargs):
        warnings.warn("rattail.util.OrderedDict is deprecated; "
                      "please use collections.OrderedDict instead",
                      DeprecationWarning, stacklevel=2)
        super(OrderedDict, self).__init__(*args, **kwargs)


def capture_output(*args, **kwargs):
    """
    Runs ``command`` and returns any output it produces.
    """
    # We *need* to pipe ``stdout`` because that's how we capture the output of
    # the ``hg`` command.  However, we must pipe *all* handles in order to
    # prevent issues when running as a GUI but *from* the Windows console.  See
    # also: http://bugs.python.org/issue3905
    kwargs.update({
        'stdin': subprocess.PIPE,
        'stdout': subprocess.PIPE,
        'stderr': subprocess.PIPE,
    })
    output = subprocess.Popen(*args, **kwargs).communicate()[0]
    return output


def data_diffs(local_data, host_data, fields=None,
               fuzzy_fields=None, fuzz_factor=None):
    """
    Find all (relevant) fields which differ between the host and local data
    values for a given record.
    """
    if fields is None:
        fields = list(local_data)

    diffs = []
    fuzzy_time_cutoff = datetime.timedelta(seconds=fuzz_factor or 1)
    for field in fields:

        # maybe raise explicit error to indicate *which* dict is missing data
        # (otherwise the comparison below raises "ambiguous" error about it)
        if field not in local_data:
            raise KeyError("key '{}' is missing from local_data".format(field))
        if field not in host_data:
            raise KeyError("key '{}' is missing from host_data".format(field))

        local_value = local_data[field]
        host_value = host_data[field]

        if fuzzy_fields and field in fuzzy_fields:

            if local_value and not host_value:
                diffs.append(field)
                continue

            if host_value and not local_value:
                diffs.append(field)
                continue

            if isinstance(local_value, datetime.datetime):

                if (host_value > local_value
                    and host_value - local_value >= fuzzy_time_cutoff):
                    diffs.append(field)

                elif (local_value > host_value
                      and local_value - host_value >= fuzzy_time_cutoff):
                    diffs.append(field)

                continue

            elif isinstance(local_value, (int, float, decimal.Decimal)):

                if (host_value > local_value
                    and host_value - local_value >= fuzz_factor):
                    diffs.append(field)

                elif (local_value > host_value
                      and local_value - host_value >= fuzz_factor):
                    diffs.append(field)

                continue

        if local_value != host_value:
            diffs.append(field)

    return diffs


def get_class_hierarchy(klass, topfirst=True):
    """
    Returns a list of all classes in the inheritance chain for the
    given class.

    For instance::

       class A(object):
          pass

       class B(A):
          pass

       class C(B):
          pass

       get_class_hierarchy(C)
       # -> [A, B, C]

    Specify ``topfirst=False`` to get ``[C, B, A]`` instead.
    """
    hierarchy = []

    def traverse(cls):
        if cls is not object:
            hierarchy.append(cls)
            for parent in cls.__bases__:
                traverse(parent)

    traverse(klass)
    if topfirst:
        hierarchy.reverse()
    return hierarchy


def get_package_name(name):
    """
    Generic logic to derive a "package name" from the given name.

    E.g. if ``name`` is "Poser Plus" this will return "poser_plus"
    """
    # cf. https://stackoverflow.com/a/3194567
    name = unicodedata.normalize('NFD', name)
    name = name.encode('ascii', 'ignore').decode('ascii')
    words = re.split(r'[\- ]', name)
    return '_'.join([word.lower() for word in words])


def get_studly_prefix(name):
    """
    Generic logic to derive a "studly prefix" from the given name.

    E.g. if ``name`` is "Poser Plus" this will return "PoserPlus"
    """
    # cf. https://stackoverflow.com/a/3194567
    name = unicodedata.normalize('NFD', name)
    name = name.encode('ascii', 'ignore').decode('ascii')
    words = re.split(r'[\- ]', name)
    return ''.join([word.capitalize() for word in words])


def import_module_path(module_path):
    """
    Import an arbitrary Python module.

    .. warning::

       This function is deprecated; please use
       :func:`python:importlib.import_module()` instead.

    :param module_path: String referencing a module by its "dotted path".

    :returns: The referenced module.
    """
    warnings.warn("rattail.util.import_module_path() is deprecated; "
                  "please use importlib.import_module() instead",
                  DeprecationWarning, stacklevel=2)

    return importlib.import_module(module_path)


def import_reload(module):
    """
    Reload a module.

    .. warning::

       This function is deprecated; please use
       :func:`python:importlib.reload()` instead.

    :param module: An already-loaded module.

    :returns: The module.
    """
    warnings.warn("rattail.util.import_reload() is deprecated; "
                  "please use importlib.reload() instead",
                  DeprecationWarning, stacklevel=2)

    return importlib.reload(module)


def get_object_spec(obj):
    """
    Returns the "object spec" string for the given object.
    """
    # TODO: this only handles a class *instance* currently
    return '{}:{}'.format(obj.__module__, obj.__class__.__name__)


def load_object(spec):
    """
    Load an arbitrary object from a module, according to a specifier.

    This is a compatibility wrapper around
    :func:`wuttjamaican:wuttjamaican.util.load_object()`.  New code
    should use that instead; this will eventually be removed.
    """
    return wutta_load_object(spec)


def load_entry_points(group, ignore_errors=False):
    """
    Load a set of ``setuptools``-style entry points.

    This is a compatibility wrapper around
    :func:`wuttjamaican:wuttjamaican.util.load_entry_points()`.  New
    code should use that instead; this will eventually be removed.
    """
    return wutta_load_entry_points(group, ignore_errors=ignore_errors)


def prettify(text):
    """
    Return a "prettified" version of text.
    """
    text = text.replace('_', ' ')
    text = text.replace('-', ' ')
    words = text.split()
    return ' '.join([x.capitalize() for x in words])


def pretty_boolean(value):
    """
    Returns ``'Yes'`` or ``'No'`` or empty string if value is ``None``
    """
    if value is None:
        return ""
    return "Yes" if value else "No"


def hours_as_decimal(hours, places=2):
    """
    Convert the given ``datetime.timedelta`` object into a Decimal whose
    value is in terms of hours.
    """
    if hours is None:
        return
    minutes = (hours.days * 1440) + (hours.seconds // 60)
    fmt = '{{:0.{}f}}'.format(places)
    return decimal.Decimal(fmt.format(minutes / 60.0))


def pretty_hours(hours=None, seconds=None):
    """ DEPRECATED """
    warnings.warn("pretty_hours() function is deprecated; please "
                  "use AppHandler.render_duration() insted",
                  DeprecationWarning, stacklevel=2)

    if hours is None and seconds is None:
        return ''
    if hours is None:
        hours = datetime.timedelta(seconds=seconds)

    # determine if hours value is positive or negative.  seems like some of the
    # math can be "off" if negative values are involved, in which case we'll
    # convert to positive for sake of math and then prefix result with a hyphen
    negative = False

    if isinstance(hours, decimal.Decimal):
        if hours < 0:
            negative = True
            hours = -hours
        hours = datetime.timedelta(seconds=int(hours * 3600))

    minutes = (hours.days * 1440) + (hours.seconds / 60)
    rendered = "{}:{:02d}".format(int(minutes // 60), int(minutes % 60))
    if negative:
        rendered = "-{}".format(rendered)
    return rendered


def pretty_quantity(value, empty_zero=False):
    """
    Return a "pretty" version of the given value, as string.  This is meant primarily
    for use with things like order quantities, so that e.g. 1.0000 => 1
    """
    if value is None:
        return ''
    if int(value) == value:
        value = int(value)
        if empty_zero and value == 0:
            return ''
        return str(value)
    return str(value).rstrip('0')


def progress_loop(func, items, factory, *args, **kwargs):
    """
    This will iterate over ``items`` and call ``func`` for each.  If a progress
    ``factory`` kwarg is provided, then a progress instance will be created and
    updated along the way.
    """
    message = kwargs.pop('message', None)
    count = kwargs.pop('count', None)
    allow_cancel = kwargs.pop('allow_cancel', False)
    if count is None:
        try:
            count = len(items)
        except TypeError:
            count = items.count()
    if not count:
        return True

    prog = None
    if factory:
        prog = factory(message, count)

    canceled = False
    for i, item in enumerate(items, 1):
        func(item, i, *args, **kwargs)
        if prog and not prog.update(i):
            canceled = True
            break
    if prog:
        prog.finish()
    if canceled and not allow_cancel:
        raise RuntimeError("Operation was canceled")
    return not canceled


def simple_error(error):
    """
    Return a "simple" string for the given error.  Result will look like::

       "ErrorClass: Description for the error"

    However the logic checks to ensure the error has a descriptive message
    first; if it doesn't the result will just be::

       "ErrorClass"
    """
    cls = str(type(error).__name__)
    msg = str(error)
    if msg:
        return "{}: {}".format(cls, msg)
    return cls


def shlex_join(split_command):
    """
    Wrapper which invokes ``shlex.join()`` if available; otherwise
    a workaround is implemented.

    cf. https://stackoverflow.com/a/50022816
    """
    # python 3.8+ has native shlex.join()
    if hasattr(shlex, 'join'):
        return shlex.join(split_command)

    # python 3.3+ has native shlex.quote()
    if hasattr(shlex, 'quote'):
        return ' '.join(shlex.quote(x) for x in split_command)

    # python 2 has neither...so the rest of this logic is copy/pasted
    # from later python
    # _find_unsafe = re.compile(r'[^\w@%+=:,./-]', re.ASCII).search
    _find_unsafe = re.compile(r'[^\w@%+=:,./-]').search

    def quote(s):
        """Return a shell-escaped version of the string *s*."""
        if not s:
            return "''"
        if _find_unsafe(s) is None:
            return s

        # use single quotes, and put single quotes into double quotes
        # the string $'b is then quoted as '$'"'"'b'
        return "'" + s.replace("'", "'\"'\"'") + "'"

    return ' '.join(quote(x) for x in split_command)



# -*- coding: utf-8; -*-

import unittest

from rattail import mail
from rattail.config import RattailConfig


class TestEmail(unittest.TestCase):

    def test_template_lookup_paths(self):

        # empty (no paths) by default
        config = RattailConfig()
        email = mail.Email(config, 'testing')
        self.assertEqual(email.html_templates.directories, [])
        
        # config may specify paths
        config = RattailConfig()
        config.setdefault('rattail.mail', 'templates', '/tmp/foo /tmp/bar')
        email = mail.Email(config, 'testing')
        self.assertEqual(email.html_templates.directories, ['/tmp/foo', '/tmp/bar'])



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
POS Batch Handler
"""

import decimal

import sqlalchemy as sa
from sqlalchemy import orm

from rattail.batch import BatchHandler
from rattail.db.model import POSBatch


class POSBatchHandler(BatchHandler):
    """
    Handler for POS batches
    """
    batch_model_class = POSBatch

    def get_terminal_id(self):
        """
        Returns the ID string for current POS terminal.
        """
        return self.config.get('rattail', 'pos.terminal_id')

    def init_batch(self, batch, **kwargs):
        batch.terminal_id = self.get_terminal_id()
        batch.status_code = batch.STATUS_OK

    def make_row(self, **kwargs):
        row = super().make_row(**kwargs)
        row.timestamp = self.app.make_utc()
        return row

    # TODO: should also filter this by terminal
    def get_current_batch(
            self,
            user,
            training_mode=False,
            create=True,
            return_created=False,
            **kwargs):
        """
        Get the "current" POS batch for the given user, creating it as
        needed.

        :param training_mode: Flag indicating whether the transaction
           should be in training mode.  The lookup will be restricted
           according to the value of this flag.  If a new batch is
           created, it will be assigned this flag value.
        """
        if not user:
            raise ValueError("must specify a user")

        created = False
        model = self.model
        session = self.app.get_session(user)
        employee = self.app.get_employee(user)

        # TODO: can we assume cashier (employee) is always set /
        # accurate?  if so maybe stop filtering on created_by?
        cashier_criteria = sa.and_(
            model.POSBatch.cashier == None,
            model.POSBatch.created_by == user)
        if employee:
            cashier_criteria = sa.or_(
                model.POSBatch.cashier == employee,
                cashier_criteria)

        try:
            batch = session.query(model.POSBatch)\
                           .filter(cashier_criteria)\
                           .filter(model.POSBatch.status_code == model.POSBatch.STATUS_OK)\
                           .filter(model.POSBatch.executed == None)\
                           .filter(model.POSBatch.training_mode == training_mode)\
                           .one()

        except orm.exc.NoResultFound:
            if not create:
                if return_created:
                    return None, False
                return
            batch = self.make_batch(session, created_by=user,
                                    training_mode=training_mode,
                                    cashier=employee)
            session.add(batch)
            session.flush()
            created = True

        if return_created:
            return batch, created

        return batch

    def get_screen_txn_display(self, batch, **kwargs):
        """
        Should return the text to be used for displaying transaction
        identifier within the header of POS screen.
        """
        return batch.id_str

    def get_screen_cust_display(self, batch=None, customer=None, **kwargs):
        """
        Should return the text to be used for displaying customer
        identifier / name etc. within the header of POS screen.
        """
        if not customer and batch:
            customer = batch.customer
        if not customer:
            return

        key_field = self.app.get_customer_key_field()
        customer_key = getattr(customer, key_field)
        return str(customer_key or '')

    # TODO: this should account for shoppers somehow too
    def set_customer(self, batch, customer, user=None, **kwargs):
        """
        Assign the customer account for POS transaction.
        """
        if customer and batch.customer:
            row_type = self.enum.POS_ROW_TYPE_SWAP_CUSTOMER
        elif customer:
            row_type = self.enum.POS_ROW_TYPE_SET_CUSTOMER
        else:
            if not batch.customer:
                return
            row_type = self.enum.POS_ROW_TYPE_DEL_CUSTOMER

        batch.customer = customer
        if customer:
            member = self.app.get_member(customer)
            batch.customer_is_member = bool(member and member.active)
            employee = self.app.get_employee(customer)
            batch.customer_is_employee = bool(employee and
                                              employee.status == self.enum.EMPLOYEE_STATUS_CURRENT)
        else:
            batch.customer_is_member = None
            batch.customer_is_employee = None

        row = self.make_row()
        row.user = user
        row.row_type = row_type
        if customer:
            key = self.app.get_customer_key_field()
            row.item_entry = getattr(customer, key)
        row.description = str(customer) if customer else 'REMOVE CUSTOMER'
        self.add_row(batch, row)

    def process_entry(self, batch, entry, quantity=1, user=None, **kwargs):
        """
        Process an "entry" value direct from POS.  Most typically,
        this is effectively "ringing up an item" and hence we add a
        row to the batch and return the row.
        """
        session = self.app.get_session(batch)
        model = self.model

        if isinstance(entry, model.Product):
            product = entry
            entry = product.uuid
        else:
            # TODO: maybe should only search by product key (or GPC?) etc.
            product = self.app.get_products_handler().locate_product_for_entry(session, entry)

        # TODO: if product not found, should auto-record a badscan
        # entry?  maybe only if config says so, e.g. might be nice to
        # only record badscan if entry truly came from scanner device,
        # in which case only the caller would know that
        if not product:
            return

        if product.not_for_sale:
            key = self.app.get_products_handler().render_product_key(product)
            raise ValueError(f"product is not for sale: {key}")

        # product located, so add item row
        row = self.make_row()
        row.user = user
        row.item_entry = kwargs.get('item_entry', entry)
        row.upc = product.upc
        row.item_id = product.item_id
        row.product = product
        row.brand_name = product.brand.name if product.brand else None
        row.description = product.description
        row.size = product.size
        row.full_description = product.full_description
        dept = product.department
        if dept:
            row.department_number = dept.number
            row.department_name = dept.name
        subdept = product.subdepartment
        if subdept:
            row.subdepartment_number = subdept.number
            row.subdepartment_name = subdept.name
        row.foodstamp_eligible = product.food_stampable
        row.sold_by_weight = product.weighed # TODO?
        row.quantity = quantity

        regprice = product.regular_price
        if regprice:
            row.reg_price = regprice.price

        curprice = product.current_price
        if curprice:
            row.cur_price = curprice.price
            row.cur_price_type = curprice.type
            row.cur_price_start = curprice.starts
            row.cur_price_end = curprice.ends

        row.txn_price = row.cur_price or row.reg_price

        if row.txn_price:
            row.sales_total = row.txn_price * row.quantity
            batch.sales_total = (batch.sales_total or 0) + row.sales_total
            if row.foodstamp_eligible:
                batch.fs_total = (batch.fs_total or 0) + row.sales_total

        tax = product.tax
        if tax:
            row.tax_code = tax.code

        if row.txn_price:
            row.row_type = self.enum.POS_ROW_TYPE_SELL
            if tax:
                self.update_tax(batch, row, tax)
        else:
            row.row_type = self.enum.POS_ROW_TYPE_BADPRICE

        self.add_row(batch, row)
        session.flush()
        return row

    def update_tax(self, batch, row, tax=None, tax_code=None, **kwargs):
        """
        Update the tax totals for the batch, basd on given row.
        """
        if not tax and not tax_code:
            raise ValueError("must specify either tax or tax_code")

        session = self.app.get_session(batch)
        if not tax:
            tax = self.get_tax(session, tax_code)

        btax = batch.taxes.get(tax.code)
        if not btax:
            btax = self.model.POSBatchTax()
            btax.tax = tax
            btax.tax_code = tax.code
            btax.tax_rate = tax.rate
            session.add(btax)
            btax.batch = batch
            session.flush()

        # calculate relevant sales
        rows = [r for r in batch.active_rows()
                if r.tax_code == tax.code
                and not r.void]
        sales = sum([r.sales_total for r in rows])
        # nb. must add row separately if not yet in batch
        if not row.batch and not row.batch_uuid:
            sales += row.sales_total

        # total for this tax
        before = btax.tax_total or 0
        btax.tax_total = (sales * (tax.rate / 100)).quantize(decimal.Decimal('0.02'))
        batch.tax_total = (batch.tax_total or 0) - before + btax.tax_total

    def record_badscan(self, batch, entry, quantity=1, user=None, **kwargs):
        """
        Add a row to the batch which represents a "bad scan" at POS.
        """
        row = self.make_row()
        row.user = user
        row.row_type = self.enum.POS_ROW_TYPE_BADSCAN
        row.item_entry = entry
        row.description = "BADSCAN"
        row.quantity = quantity
        self.add_row(batch, row)
        return row

    def add_open_ring(self, batch, department, price, quantity=1, user=None, **kwargs):
        """
        Adds an "open ring" row to the batch.
        """
        session = self.app.get_session(batch)
        model = self.model

        if not isinstance(department, model.Department):
            department = session.get(model.Department, department)
            if not department:
                raise ValueError("must specify valid department")

        # add row for open ring
        row = self.make_row()
        row.row_type = self.enum.POS_ROW_TYPE_OPEN_RING
        row.user = user
        row.item_entry = department.number
        row.description = f"OPEN RING: {department.name}"
        row.department_number = department.number
        row.department_name = department.name
        row.foodstamp_eligible = department.food_stampable
        row.quantity = quantity

        row.txn_price = price
        if row.txn_price:
            row.sales_total = row.txn_price * row.quantity
            batch.sales_total = (batch.sales_total or 0) + row.sales_total
            if row.foodstamp_eligible:
                batch.fs_total = (batch.fs_total or 0) + row.sales_total

        tax = department.tax
        if tax:
            row.tax_code = tax.code
            self.update_tax(batch, row, tax)

        self.add_row(batch, row)
        session.flush()
        return row

    def get_tax(self, session, code, **kwargs):
        """
        Return the tax record corresponding to the given code.

        :param session: Current DB session.

        :param code: Tax code to fetch.
        """
        model = self.model
        return session.query(model.Tax)\
                      .filter(model.Tax.code == code)\
                      .one()

    def get_tender(self, session, key, **kwargs):
        """
        Return the tender record corresponding to the given key.

        :param session: Current DB session.

        :param key: Either a tender UUID, or "true" tender code (i.e.
           :attr:`rattail.db.model.sales.Tender.code` value) or a
           "pseudo-code" for common tenders (e.g. ``'cash'``).
        """
        model = self.model

        # Tender.uuid match?
        tender = session.get(model.Tender, key)
        if tender:
            return tender

        # Tender.code match?
        try:
            return session.query(model.Tender)\
                          .filter(model.Tender.code == key)\
                          .one()
        except orm.exc.NoResultFound:
            pass

        # try settings, if value then recurse
        # TODO: not sure why get_vendor() only checks settings?
        # for now am assuming we should also check config file
        #key = self.app.get_setting(session, f'rattail.tender.{key}')
        key = self.config.get('rattail', f'tender.{key}')
        if key:
            return self.get_tender(session, key, **kwargs)

    def refresh_row(self, row):
        # TODO (?)
        row.status_code = row.STATUS_OK
        row.status_text = None

    def clone(self, oldbatch, created_by, **kwargs):
        newbatch = super().clone(oldbatch, created_by, **kwargs)
        session = self.app.get_session(oldbatch)
        model = self.model

        tax_mapper = sa.inspect(model.POSBatchTax)
        for oldtax in oldbatch.taxes.values():
            newtax = model.POSBatchTax()
            for key in tax_mapper.columns.keys():
                if key not in ('uuid', 'batch_uuid'):
                    setattr(newtax, key, getattr(oldtax, key))
            session.add(newtax)
            newtax.batch = newbatch

        return newbatch

    def get_clonable_rows(self, batch, **kwargs):
        # TODO: row types..ugh
        return [row for row in batch.data_rows
                if row.row_type != self.enum.POS_ROW_TYPE_TENDER]

    def override_price(self, row, user, txn_price, **kwargs):
        """
        Override the transaction price for the given batch row.
        """
        batch = row.batch

        # update price for given row
        orig_row = row
        orig_txn_price = orig_row.txn_price
        orig_sales_total = orig_row.sales_total
        orig_row.txn_price = txn_price
        orig_row.txn_price_adjusted = True
        orig_row.sales_total = orig_row.quantity * orig_row.txn_price

        # adjust totals
        batch.sales_total = (batch.sales_total or 0) - orig_sales_total + orig_row.sales_total
        if orig_row.foodstamp_eligible:
            batch.fs_total = (batch.fs_total or 0) - orig_sales_total + orig_row.sales_total
        if orig_row.tax_code:
            self.update_tax(batch, orig_row, tax_code=orig_row.tax_code)

        # add another row indicating who/when
        row = self.make_row()
        row.user = user
        row.row_type = self.enum.POS_ROW_TYPE_ADJUST_PRICE
        row.item_entry = orig_row.item_entry
        row.txn_price = txn_price
        row.description = (f"ROW {orig_row.sequence} PRICE ADJUST "
                           f"FROM {self.app.render_currency(orig_txn_price)}")
        self.add_row(batch, row)
        return row

    def void_row(self, row, user, **kwargs):
        """
        Apply "void" status to the given batch row.
        """
        batch = row.batch

        # mark given row as void
        orig_row = row
        orig_row.void = True

        # adjust batch totals
        if orig_row.sales_total:
            batch.sales_total = (batch.sales_total or 0) - orig_row.sales_total
            if orig_row.foodstamp_eligible:
                batch.fs_total = (batch.fs_total or 0) - orig_row.sales_total
        if orig_row.tax_code:
            self.update_tax(batch, orig_row, tax_code=orig_row.tax_code)

        # add another row indicating who/when
        row = self.make_row()
        row.user = user
        row.row_type = self.enum.POS_ROW_TYPE_VOID_LINE
        row.item_entry = orig_row.item_entry
        row.description = f"VOID ROW {orig_row.sequence}"
        self.add_row(batch, row)
        return row

    def suspend_transaction(self, batch, user, **kwargs):
        """
        Suspend transaction for the given POS batch.
        """
        # add another row indicating who/when
        row = self.make_row()
        row.user = user
        row.row_type = self.enum.POS_ROW_TYPE_SUSPEND
        row.description = "SUSPEND TXN"
        self.add_row(batch, row)

        # TODO: should do something different if we have a server
        # engine, i.e. central location for suspended txns

        # mark batch as suspended (but not executed)
        batch.status_code = batch.STATUS_SUSPENDED

    def resume_transaction(self, batch, user, **kwargs):
        """
        Resume transaction for the given POS batch.  By default this
        always creates a *new* batch.
        """
        # TODO: should do something different if we have a server
        # engine, i.e. central location for suspended txns

        newbatch = self.clone(batch, user)
        newbatch.cashier = self.app.get_employee(user)
        newbatch.status_code = newbatch.STATUS_OK

        # add another row indicating who/when
        row = self.make_row()
        row.user = user
        row.row_type = self.enum.POS_ROW_TYPE_RESUME
        row.description = "RESUME TXN"
        self.add_row(newbatch, row)

        # mark original batch as executed
        batch.executed = self.app.make_utc()
        batch.executed_by = user

        return newbatch

    def void_batch(self, batch, user, **kwargs):
        """
        Void the given POS batch.
        """
        # add another row indicating who/when
        row = self.make_row()
        row.user = user
        row.row_type = self.enum.POS_ROW_TYPE_VOID_TXN
        row.description = "VOID TXN"
        self.add_row(batch, row)

        # void/execute batch
        batch.void = True
        batch.executed = self.app.make_utc()
        batch.executed_by = user

    def apply_tender(self, batch, user, tender, amount, **kwargs):
        """
        Apply the given tender amount to the batch.

        :param tender: Reference to a
           :class:`~rattail.db.model.sales.Tender` or similar object, or dict
           with similar keys, or can be just a tender code.

        :param amount: Amount to apply.  Note, this usually should be
           a *negative* number.

        :returns: List of rows which were added to the batch.
        """
        session = self.app.get_session(batch)
        model = self.model

        tender_info = tender
        tender = None
        item_entry = None
        description = None

        if isinstance(tender_info, model.Tender):
            tender = tender_info
        else:
            if isinstance(tender_info, str):
                tender_code = tender_info
            else:
                tender_code = tender_info['code']
            item_entry = tender_code
            description = f"TENDER '{tender_code}'"
            tender = self.get_tender(session, tender_code)

        if tender:
            item_entry = tender.code
            description = tender.name

            if tender.disabled:
                # TODO: behavior here should be configurable, probably
                # needs a dedicated email etc. ..or maybe just ignore
                log.error("disabled tender '%s' being applied to POS batch: %s",
                          tender, batch)

        rows = []

        # add row for tender
        row = self.make_row()
        row.user = user
        row.row_type = self.enum.POS_ROW_TYPE_TENDER
        row.item_entry = item_entry
        row.description = description
        row.tender_total = amount
        row.tender = tender
        batch.tender_total = (batch.tender_total or 0) + row.tender_total
        if tender and tender.is_foodstamp:
            batch.fs_tender_total = (batch.fs_tender_total or 0) + row.tender_total
        self.add_row(batch, row)
        rows.append(row)

        # nothing more to do for now, if balance remains
        balance = batch.get_balance()
        if balance > 0:
            return rows

        # next we'll give change back
        # nb. if balance is 0, then change due is 0, but we always
        # include the change due line item even so..

        # ..but some tenders do not allow cash back
        if balance < 0:
            if hasattr(tender, 'is_cash') and not tender.is_cash:
                if not tender.allow_cash_back:
                    raise ValueError(f"tender '{tender.name}' does not allow "
                                     f" cash back: ${-balance:0.2f}")

        # TODO: maybe should always give change as 'cash'
        if tender and tender.is_cash:
            cash = tender
        else:
            cash = self.get_tender(session, 'cash')

        row = self.make_row()
        row.user = user
        row.row_type = self.enum.POS_ROW_TYPE_CHANGE_BACK
        row.item_entry = item_entry
        row.description = "CHANGE DUE"
        row.tender_total = -balance
        row.tender = cash
        batch.tender_total = (batch.tender_total or 0) + row.tender_total
        self.add_row(batch, row)
        rows.append(row)

        # all paid up, so finalize
        session.flush()
        assert batch.get_balance() == 0
        self.do_execute(batch, user, **kwargs)
        return rows

    def execute(self, batch, progress=None, **kwargs):
        # TODO
        return True



# -*- coding: utf-8; -*-
"""add MemberEquityPayment.source

Revision ID: a7c2d5d84743
Revises: 310080aaa346
Create Date: 2023-09-07 17:33:19.223569

"""

# revision identifiers, used by Alembic.
revision = 'a7c2d5d84743'
down_revision = '310080aaa346'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # member_equity_payment
    op.add_column('member_equity_payment', sa.Column('source', sa.String(length=50), nullable=True))
    op.add_column('member_equity_payment_version', sa.Column('source', sa.String(length=50), autoincrement=False, nullable=True))


def downgrade():

    # member_equity_payment
    op.drop_column('member_equity_payment_version', 'source')
    op.drop_column('member_equity_payment', 'source')



# -*- coding: utf-8 -*-
"""add inventory reason code

Revision ID: 2fe69ea73233
Revises: 5b88c82dbd8d
Create Date: 2017-07-11 21:30:01.552398

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '2fe69ea73233'
down_revision = u'5b88c82dbd8d'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_inventory
    op.add_column('batch_inventory', sa.Column('reason_code', sa.String(length=50), nullable=True))


def downgrade():

    # batch_inventory
    op.drop_column('batch_inventory', 'reason_code')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
``rattail filemon`` command
"""

import sys
from enum import Enum
from pathlib import Path

import typer
from typing_extensions import Annotated

from .base import rattail_typer, Subcommand


class ServiceAction(str, Enum):
    start = 'start'
    stop = 'stop'
    install = 'install'
    debug = 'debug'
    uninstall = 'uninstall'


@rattail_typer.command()
def filemon(
        ctx: typer.Context,
        action: Annotated[
            ServiceAction,
            typer.Argument(help="Action to perform for the service.  Note that only "
                           "start/stop are used on Linux; the rest are Windows-only.")] = ...,
        pidfile: Annotated[
            Path,
            typer.Option('--pidfile', '-p',
                         help="Path to PID file.  (Only used on Linux.)")] = None,
        daemonize: Annotated[
            bool,
            typer.Option(help="DEPRECATED")] = False,
        username: Annotated[
            str,
            typer.Option(help="User account under which the service should run.  "
                         "(Only used on Windows.)")] = None,
        auto_start: Annotated[
            bool,
            typer.Option('--auto-start', '-a',
                         help="Configure service to start automatically.  "
                         "(Only used on Windows.)")] = False,
):
    """
    Manage the file monitor daemon
    """
    config = ctx.parent.rattail_config

    if sys.platform in ('linux', 'linux2'):
        from rattail.filemon import linux as filemon

        if action == 'start':
            filemon.start_daemon(config, pidfile)

        elif action == 'stop':
            filemon.stop_daemon(config, pidfile)

    elif sys.platform == 'win32': # pragma no cover
        run_win32(ctx.params)

    else:
        sys.stderr.write(f"File monitor is not supported on platform: {sys.platform}\n")
        sys.exit(1)


class FileMonitorCommand(Subcommand):
    """
    Interacts with the file monitor service; called as ``rattail filemon``.
    This command expects a subcommand; one of the following:

    * ``rattail filemon start``
    * ``rattail filemon stop``

    On Windows platforms, the following additional subcommands are available:

    * ``rattail filemon install``
    * ``rattail filemon uninstall`` (or ``rattail filemon remove``)

    .. note::
       The Windows Vista family of operating systems requires you to launch
       ``cmd.exe`` as an Administrator in order to have sufficient rights to
       run the above commands.

    .. See :doc:`howto.use_filemon` for more information.
    """
    name = 'filemon'
    description = "Manage the file monitor daemon"

    def add_parser_args(self, parser):
        subparsers = parser.add_subparsers(title='subcommands')

        start = subparsers.add_parser('start', help="Start service")
        start.set_defaults(subcommand='start')
        stop = subparsers.add_parser('stop', help="Stop service")
        stop.set_defaults(subcommand='stop')

        if sys.platform in ('linux', 'linux2'):
            parser.add_argument('-p', '--pidfile',
                                help="Path to PID file.", metavar='PATH')
            parser.add_argument('--daemonize', action='store_true',
                                help="DEPRECATED")
            parser.add_argument('--no-daemonize',
                                '-D', '--do-not-daemonize',
                                action='store_false', dest='daemonize',
                                help="DEPRECATED")

        elif sys.platform == 'win32': # pragma no cover

            install = subparsers.add_parser('install', help="Install service")
            install.set_defaults(subcommand='install')
            install.add_argument('-a', '--auto-start', action='store_true',
                                 help="Configure service to start automatically.")
            install.add_argument('-U', '--username',
                                 help="User account under which the service should run.")

            remove = subparsers.add_parser('remove', help="Uninstall (remove) service")
            remove.set_defaults(subcommand='remove')

            uninstall = subparsers.add_parser('uninstall', help="Uninstall (remove) service")
            uninstall.set_defaults(subcommand='remove')

            debug = subparsers.add_parser('debug', help="Run service in debug mode")
            debug.set_defaults(subcommand='debug')

    def run(self, args):
        if sys.platform in ('linux', 'linux2'):
            from rattail.filemon import linux as filemon

            if args.subcommand == 'start':
                filemon.start_daemon(self.config, args.pidfile)

            elif args.subcommand == 'stop':
                filemon.stop_daemon(self.config, args.pidfile)

        elif sys.platform == 'win32': # pragma no cover
            params = dict(args._get_kwargs())
            params['action'] = args.subcommand
            run_win32(params)

        else:
            self.stderr.write("File monitor is not supported on platform: {0}\n".format(sys.platform))
            sys.exit(1)


def run_win32(params): # pragma no cover
    from rattail.win32 import require_elevation
    from rattail.win32 import service
    from rattail.win32 import users
    from rattail.filemon import win32 as filemon

    require_elevation()

    options = []
    if params['action'] == 'install':

        username = params['username']
        if username:
            if '\\' in username:
                server, username = username.split('\\')
            else:
                server = socket.gethostname()
            if not users.user_exists(username, server):
                sys.stderr.write("User does not exist: {0}\\{1}\n".format(server, username))
                sys.exit(1)

            password = ''
            while password == '':
                password = getpass(str("Password for service user: ")).strip()
            options.extend(['--username', r'{0}\{1}'.format(server, username)])
            options.extend(['--password', password])

        if params['auto_start']:
            options.extend(['--startup', 'auto'])

    service.execute_service_command(filemon, params['action'], *options)

    # If installing with custom user, grant "logon as service" right.
    if params['action'] == 'install' and params['username']:
        users.allow_logon_as_service(username)

    # TODO: Figure out if the following is even required, or if instead we
    # should just be passing '--startup delayed' to begin with?

    # If installing auto-start service on Windows 7, we should update
    # its startup type to be "Automatic (Delayed Start)".
    # TODO: Improve this check to include Vista?
    if params['action'] == 'install' and params['auto_start']:
        if platform.release() == '7':
            name = filemon.RattailFileMonitor._svc_name_
            service.delayed_auto_start_service(name)



# -*- coding: utf-8; -*-
"""fix vendor sample file version table

Revision ID: 0aeaac17cd6e
Revises: 524373deb98e
Create Date: 2023-05-08 21:35:25.611702

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '0aeaac17cd6e'
down_revision = '524373deb98e'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # vendor_sample_file
    op.alter_column('vendor_sample_file_version', 'vendor_uuid',
               existing_type=sa.VARCHAR(length=32),
               nullable=True)
    op.alter_column('vendor_sample_file_version', 'file_type',
               existing_type=sa.VARCHAR(length=100),
               nullable=True)
    op.alter_column('vendor_sample_file_version', 'filename',
               existing_type=sa.VARCHAR(length=100),
               nullable=True)
    op.alter_column('vendor_sample_file_version', 'created_by_uuid',
               existing_type=sa.VARCHAR(length=32),
               nullable=True)


def downgrade():

    # vendor_sample_file
    op.alter_column('vendor_sample_file_version', 'created_by_uuid',
               existing_type=sa.VARCHAR(length=32),
               nullable=False)
    op.alter_column('vendor_sample_file_version', 'filename',
               existing_type=sa.VARCHAR(length=100),
               nullable=False)
    op.alter_column('vendor_sample_file_version', 'file_type',
               existing_type=sa.VARCHAR(length=100),
               nullable=False)
    op.alter_column('vendor_sample_file_version', 'vendor_uuid',
               existing_type=sa.VARCHAR(length=32),
               nullable=False)



# -*- coding: utf-8; -*-
"""add POSBatch

Revision ID: 6dd836b9829a
Revises: fa3aec1556bc
Create Date: 2023-09-23 09:10:07.049227

"""

# revision identifiers, used by Alembic.
revision = '6dd836b9829a'
down_revision = 'fa3aec1556bc'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_pos
    op.create_table('batch_pos',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('description', sa.String(length=255), nullable=True),
                    sa.Column('created', sa.DateTime(), nullable=False),
                    sa.Column('created_by_uuid', sa.String(length=32), nullable=False),
                    sa.Column('cognized', sa.DateTime(), nullable=True),
                    sa.Column('cognized_by_uuid', sa.String(length=32), nullable=True),
                    sa.Column('rowcount', sa.Integer(), nullable=True),
                    sa.Column('complete', sa.Boolean(), nullable=False),
                    sa.Column('executed', sa.DateTime(), nullable=True),
                    sa.Column('executed_by_uuid', sa.String(length=32), nullable=True),
                    sa.Column('purge', sa.Date(), nullable=True),
                    sa.Column('notes', sa.Text(), nullable=True),
                    sa.Column('params', rattail.db.types.JSONTextDict(), nullable=True),
                    sa.Column('extra_data', sa.Text(), nullable=True),
                    sa.Column('status_code', sa.Integer(), nullable=True),
                    sa.Column('status_text', sa.String(length=255), nullable=True),
                    sa.Column('store_id', sa.String(length=10), nullable=True),
                    sa.Column('store_uuid', sa.String(length=32), nullable=True),
                    sa.Column('start_time', sa.DateTime(), nullable=True),
                    sa.Column('customer_uuid', sa.String(length=32), nullable=True),
                    sa.Column('sales_total', sa.Numeric(precision=9, scale=2), nullable=True),
                    sa.Column('tax1_total', sa.Numeric(precision=9, scale=2), nullable=True),
                    sa.Column('tax2_total', sa.Numeric(precision=9, scale=2), nullable=True),
                    sa.Column('void', sa.Boolean(), nullable=False),
                    sa.ForeignKeyConstraint(['cognized_by_uuid'], ['user.uuid'], name='batch_pos_fk_cognized_by'),
                    sa.ForeignKeyConstraint(['created_by_uuid'], ['user.uuid'], name='batch_pos_fk_created_by'),
                    sa.ForeignKeyConstraint(['customer_uuid'], ['customer.uuid'], name='batch_pos_fk_customer'),
                    sa.ForeignKeyConstraint(['executed_by_uuid'], ['user.uuid'], name='batch_pos_fk_executed_by'),
                    sa.ForeignKeyConstraint(['store_uuid'], ['store.uuid'], name='batch_pos_fk_store'),
                    sa.PrimaryKeyConstraint('uuid')
                    )

    # batch_pos_row
    op.create_table('batch_pos_row',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('batch_uuid', sa.String(length=32), nullable=False),
                    sa.Column('sequence', sa.Integer(), nullable=False),
                    sa.Column('status_code', sa.Integer(), nullable=True),
                    sa.Column('status_text', sa.String(length=255), nullable=True),
                    sa.Column('modified', sa.DateTime(), nullable=True),
                    sa.Column('removed', sa.Boolean(), nullable=False),
                    sa.Column('row_type', sa.String(length=32), nullable=True),
                    sa.Column('item_entry', sa.String(length=32), nullable=True),
                    sa.Column('product_uuid', sa.String(length=32), nullable=True),
                    sa.Column('description', sa.String(length=60), nullable=True),
                    sa.Column('quantity', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('reg_price', sa.Numeric(precision=8, scale=3), nullable=True),
                    sa.Column('txn_price', sa.Numeric(precision=8, scale=3), nullable=True),
                    sa.Column('sales_total', sa.Numeric(precision=9, scale=2), nullable=True),
                    sa.Column('tax1_total', sa.Numeric(precision=9, scale=2), nullable=True),
                    sa.Column('tax2_total', sa.Numeric(precision=9, scale=2), nullable=True),
                    sa.ForeignKeyConstraint(['batch_uuid'], ['batch_pos.uuid'], name='batch_pos_row_fk_batch_uuid'),
                    sa.ForeignKeyConstraint(['product_uuid'], ['product.uuid'], name='batch_pos_row_fk_item'),
                    sa.PrimaryKeyConstraint('uuid')
                    )


def downgrade():

    # batch_pos*
    op.drop_table('batch_pos_row')
    op.drop_table('batch_pos')



# -*- coding: utf-8; -*-

from __future__ import unicode_literals, absolute_import

import os
from unittest import TestCase

import openpyxl

from rattail import excel as mod
from rattail.config import make_config


class TestExcelReaderXLSX(TestCase):

    def setUp(self):
        self.config = self.make_config()

    def make_config(self):
        return make_config([], extend=False)

    def test_strip_fieldnames(self):
        app = self.config.get_app()
        path = app.make_temp_file(suffix='.xlsx')

        # first make a workbook which has whitespace in column headers
        workbook = openpyxl.Workbook()
        sheet = workbook.active
        sheet.append([' first ', 'second   '])
        workbook.save(path)

        # reader should strip fieldnames by default
        reader = mod.ExcelReaderXLSX(path)
        self.assertEqual(reader.fields, ['first', 'second'])

        # unless we say not to strip them
        reader = mod.ExcelReaderXLSX(path, strip_fieldnames=False)
        self.assertEqual(reader.fields, [' first ', 'second   '])

        os.remove(path)



# -*- coding: utf-8; -*-
"""add item codes for pricing batch

Revision ID: 5dee2f796f25
Revises: 727c63fcde2d
Create Date: 2019-04-18 18:11:32.100480

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '5dee2f796f25'
down_revision = '727c63fcde2d'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_pricing_row
    op.add_column('batch_pricing_row', sa.Column('alternate_code', sa.String(length=20), nullable=True))
    op.add_column('batch_pricing_row', sa.Column('family_code', sa.Integer(), nullable=True))
    op.add_column('batch_pricing_row', sa.Column('report_code', sa.Integer(), nullable=True))
    op.add_column('batch_pricing_row', sa.Column('vendor_item_code', sa.String(length=20), nullable=True))


def downgrade():

    # batch_pricing_row
    op.drop_column('batch_pricing_row', 'vendor_item_code')
    op.drop_column('batch_pricing_row', 'report_code')
    op.drop_column('batch_pricing_row', 'family_code')
    op.drop_column('batch_pricing_row', 'alternate_code')



# -*- coding: utf-8; -*-
"""add workorder.estimated_total

Revision ID: 7d009a925f21
Revises: a20208a41889
Create Date: 2022-08-30 21:00:34.351167

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '7d009a925f21'
down_revision = 'a20208a41889'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # workorder
    op.add_column('workorder', sa.Column('estimated_total', sa.Numeric(precision=9, scale=2), nullable=True))


def downgrade():

    # workorder
    op.drop_column('workorder', 'estimated_total')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Trainwreck Handler
"""

import warnings
from collections import OrderedDict

from rattail.app import GenericHandler
from rattail.trainwreck.db import Session as TrainwreckSession


class TrainwreckHandler(GenericHandler):
    """
    Handler for Trainwreck data and databases.
    """

    def use_rotation(self):
        """
        Returns boolean indicating whether rotation should be used,
        for the Trainwreck DB(s).
        """
        return self.config.getbool('trainwreck', 'use_rotation',
                                   default=False)
    
    def current_years(self):
        """
        Returns the number of years which should be kept in the "current"
        Trainwreck DB.

        Note that this refers to the "max" number of years.  Rotation
        is done on a yearly basis, so for instance if the number of
        years to keep current is 2 (the default), then the actual
        amount of data in the current DB will vary between 1 years
        (immediately after Jan 1 pruning) and 2 years (immediately
        before Jan 1).

        In other words if today is Jan 2, and pruning has already
        occurred, then current DB would have all of last year plus
        only a day or two from this year.  But by the end of this year
        it will (still) have all of last year, plus all of this year.
        """
        return self.config.getint('trainwreck', 'current_years',
                                  default=2)

    def make_session(self, dbkey='default', **kwargs):
        """
        Make a session for a Trainwreck DB.
        """
        if 'bind' not in kwargs:
            engine = self.config.trainwreck_engines[dbkey]
            kwargs['bind'] = engine
        return TrainwreckSession(**kwargs)

    def get_model(self, **kwargs):
        """
        Return the data model for Trainwreck.
        """
        return self.config.get_trainwreck_model()

    def get_trainwreck_engines(self, include_hidden=True):
        """
        Return an "ordered" dict with configured trainwreck DB
        engines.  Keys of the dict will correspond to the config keys
        for each DB, values will be the engines.

        :param include_hidden: Flag indicating whether the result
           should include engines which are marked as hidden.  Note
           that hidden engines *are* included by default.

        :returns: An :class:`~rattail.util.OrderedDict` instance.
        """
        engines = OrderedDict(self.config.trainwreck_engines)

        if not include_hidden:
            for key in self.get_hidden_engine_keys():
                engines.pop(key, None)

        return engines

    def get_hidden_engine_keys(self):
        """
        Return a list of database engine keys which are configured to
        be hidden from the user interface.
        """
        hidden = self.config.getlist('trainwreck.db', 'hide',
                                     default=None)
        if hidden is None:
            hidden = self.config.getlist('tailbone', 'engines.trainwreck.hidden', 
                                         default=None)
            if hidden is not None:
                warnings.warn("[tailbone] 'engines.trainwreck.hidden' is a "
                              "deprecated setting, please use "
                              "[trainwreck.db] 'hide' instead",
                              DeprecationWarning, stacklevel=2)

        return hidden or []

    def engine_is_hidden(self, key):
        """
        Returns a boolean indicating if the given Trainwreck database
        engine is configured to be hidden from the user interface.
        """
        hidden = self.get_hidden_engine_keys()
        return key in hidden

    def get_oldest_transaction_date(self, session):
        """
        Query a Trainwreck database to determine the date of the
        "oldest" transaction it contains.

        :param session: SQLAlchemy session for a Trainwreck database.

        :return: A :class:`python:datetime.date` instance representing
           the oldest transaction date contained by the database.
        """
        trainwreck = self.config.get_trainwreck_model()
        txn = session.query(trainwreck.Transaction)\
                     .order_by(trainwreck.Transaction.end_time)\
                     .first()
        if txn:
            return self.app.localtime(txn.end_time, from_utc=True).date()

    def get_newest_transaction_date(self, session):
        """
        Query a Trainwreck database to determine the date of the
        "newest" transaction it contains.

        :param session: SQLAlchemy session for a Trainwreck database.

        :return: A :class:`python:datetime.date` instance representing
           the newest transaction date contained by the database.
        """
        trainwreck = self.config.get_trainwreck_model()
        txn = session.query(trainwreck.Transaction)\
                     .order_by(trainwreck.Transaction.end_time.desc())\
                     .first()
        if txn:
            return self.app.localtime(txn.end_time, from_utc=True).date()

    def prune_date(self, session, date, start_time, end_time):
        """
        Your custom handler must implement the logic for this method.
        It is responsible for fully pruning data from the given
        session's database, for the given date.  The start time and
        end times given may be used for your queries; they are in
        localtime and are zone-aware.
        """
        raise NotImplementedError



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
DataSync Configuration
"""

import re
import logging
import warnings

from rattail.config import ConfigProfile
from rattail.exceptions import ConfigurationError

from rattail.datasync.watchers import NullWatcher


log = logging.getLogger(__name__)


class DataSyncProfile(ConfigProfile):
    """
    Simple class to hold configuration for a DataSync "profile".  Each profile
    determines which database(s) will be watched for new changes, and which
    consumer(s) will then be instructed to process the changes.

    .. todo::
       This clearly needs more documentation.
    """
    section = 'rattail.datasync'

    def __init__(self, *args, **kwargs):
        load_disabled_consumers = kwargs.pop('load_disabled_consumers', False)
        super().__init__(*args, **kwargs)

        self.watcher_spec = self._config_string('watcher.spec')
        if not self.watcher_spec:
            # TODO: remove this fallback someday, since it won't work
            # with newer config logic
            self.watcher_spec = self._config_string('watcher', ignore_ambiguous=True)
            if self.watcher_spec:
                warnings.warn(f"URGENT: instead of '{self.section}.{self.prefix}.watcher', "
                              f"you should set '{self.section}.{self.prefix}.watcher.spec'",
                              DeprecationWarning, stacklevel=2)
            else:
                raise RuntimeError("no watcher spec defined")

        if self.watcher_spec == 'null':
            self.watcher = NullWatcher(self.config, self.key)
            self.watcher_kwargs = {}

        else:
            kwargs = {}

            # additional watcher kwargs will be read *directly* from
            # either the config file, or DB settings table
            handler = self.app.get_datasync_handler()
            if handler.should_use_profile_settings():

                # DB settings it is
                model = self.model
                with self.app.short_session() as s:
                    settings = s.query(model.Setting)\
                                .filter(model.Setting.name.like('rattail.datasync.{}.watcher.kwarg.%'.format(self.key)))\
                                .all()
                    for setting in settings:
                        name = setting.name.split('watcher.kwarg.')[1]
                        kwargs[name] = setting.value

            else: # config file

                pattern = re.compile(r'^{}\.watcher\.kwarg\.(?P<keyword>\w+)$'.format(self.key))
                settings = self.config.get_dict(self.section)
                for key in settings:
                    match = pattern.match(key)
                    if match:
                        keyword = match.group('keyword')
                        kwargs[keyword] = settings[key]

            # only capture "explicitly configured" watcher kwargs..
            # (nb. this is used for configuration of the kwargs)
            self.watcher_kwargs = dict(kwargs)

            # ..whereas in practice we also add one for the dbkey
            dbkey = self._config_string('watcher.db', default='default')
            kwargs['dbkey'] = dbkey

            # make the watcher
            factory = self.app.load_object(self.watcher_spec)
            self.watcher = factory(self.config, self.key, **kwargs)

        self.watcher.delay = self._config_int('watcher.delay', default=self.watcher.delay)
        self.watcher.retry_attempts = self._config_int('watcher.retry_attempts', default=self.watcher.retry_attempts)
        self.watcher.retry_delay = self._config_int('watcher.retry_delay', default=self.watcher.retry_delay)
        self.watcher.default_runas = self._config_string('consumers.runas')

        consumers = self._config_list('consumers.list')
        if not consumers:
            consumers = self._config_list('consumers', ignore_ambiguous=True)
            if consumers:
                warnings.warn(f"URGENT: instead of 'rattail.datasync.{self.key}.consumers', "
                              f"you should set 'rattail.datasync.{self.key}.consumers.list'",
                              DeprecationWarning, stacklevel=2)

        if consumers == ['self']:
            self.watcher.consumes_self = True
        else:
            self.watcher.consumes_self = False
            self.consumer_delay = self._config_int('consumer_delay', default=1)
            self.consumers = self.normalize_consumers(self.watcher.default_runas,
                                                      include_disabled=load_disabled_consumers)
            self.watcher.consumer_stub_keys = [c.key for c in self.consumers]

    def normalize_consumers(self, default_runas, include_disabled=False):
        consumers = []
        if include_disabled:
            enabled = get_consumer_keys(self.config, self.key,
                                        include_disabled=False)
        for key in get_consumer_keys(self.config, self.key,
                                     include_disabled=include_disabled):

            consumer_spec = self._config_string(f'consumer.{key}.spec')
            if not consumer_spec:
                consumer_spec = self._config_string(f'consumer.{key}', ignore_ambiguous=True)
                if consumer_spec:
                    warnings.warn(f"URGENT: instead of '{self.section}.{self.prefix}.consumer.{key}', "
                                  f"you should set '{self.section}.{self.prefix}.consumer.{key}.spec'",
                                  DeprecationWarning, stacklevel=2)
                else:
                    raise RuntimeError(f"must define '{self.section}.{self.prefix}.consumer.{key}.spec'")

            if consumer_spec == 'null':
                consumer_spec = 'rattail.datasync.consumers:NullTestConsumer'

            dbkey = self._config_string('consumer.{}.db'.format(key),
                                        default='default')
            runas = self._config_string('consumer.{}.runas'.format(key))
            try:
                factory = self.app.load_object(consumer_spec)
                consumer = factory(self.config, key, dbkey=dbkey,
                                   runas=runas or default_runas,
                                   watcher=self.watcher)
            except:
                log.debug("failed to load '%s' consumer for '%s' profile",
                          key, self.key, exc_info=True)
                if not include_disabled:
                    raise
            else:
                consumer.spec = consumer_spec
                consumer.delay = self._config_int(
                    'consumer.{}.delay'.format(key),
                    default=self.consumer_delay)
                consumer.retry_attempts = self._config_int(
                    'consumer.{}.retry_attempts'.format(key),
                    default=consumer.retry_attempts)
                consumer.retry_delay = self._config_int(
                    'consumer.{}.retry_delay'.format(key),
                    default=consumer.retry_delay)
                if include_disabled:
                    consumer.enabled = key in enabled
                consumers.append(consumer)
        return consumers


def get_consumer_keys(config, profile_key, include_disabled=False):

    # start with the primary set of consumer keys
    keys = config.getlist('rattail.datasync', f'{profile_key}.consumers.list')
    if not keys:
        keys = config.getlist('rattail.datasync', f'{profile_key}.consumers',
                              ignore_ambiguous=True)
        if keys:
            warnings.warn(f"URGENT: instead of 'rattail.datasync.{profile_key}.consumers', "
                          f"you should set 'rattail.datasync.{profile_key}.consumers.list'",
                          DeprecationWarning, stacklevel=2)
        else:
            keys = []

    if include_disabled:

        # first look in config file options
        settings = config.get_dict('rattail.datasync')
        pattern = re.compile(r'^{}\.consumer\.([^.]+)\.spec$'.format(profile_key))
        for key in settings:

            # find all consumers with spec defined
            match = pattern.match(key)
            if match:
                keys.append(match.group(1))

        # maybe also look for config settings in DB
        if config.usedb:
            app = config.get_app()
            model = config.get_model()
            session = app.make_session()
            settings = session.query(model.Setting)\
                              .filter(model.Setting.name.like(f'rattail.datasync.{profile_key}.consumer.%.spec'))\
                              .all()
            pattern = re.compile(r'^rattail\.datasync\.{}\.consumer\.(.+)\.spec$')
            for setting in settings:
                match = pattern.match(setting.name)
                if match:
                    keys.append(match.group(1))
            session.close()

    return list(sorted(set(keys)))


def get_profile_keys(config, include_disabled=False):
    """
    Returns a list of profile keys used in the DataSync configuration.
    """
    # start with the primary set of watcher keys
    keys = config.getlist('rattail.datasync', 'watch',
                          default=[])

    if include_disabled:

        # first look in config file options
        settings = config.get_dict('rattail.datasync')
        pattern = re.compile(r'^(\S+)\.watcher\.spec$')
        for key in settings:

            # find all profiles with watcher defined
            match = pattern.match(key)
            if match:
                keys.append(match.group(1))

        # maybe also look for config settings in DB
        if config.usedb:
            app = config.get_app()
            model = config.get_model()
            session = app.make_session()
            settings = session.query(model.Setting)\
                              .filter(model.Setting.name.like('rattail.datasync.%.watcher.spec'))\
                              .all()
            for setting in settings:
                parts = setting.name.split('.')
                keys.append('.'.join(parts[2:-2]))
            session.close()

    return list(sorted(set(keys)))


def load_profiles(config, include_disabled=False, ignore_problems=False):
    """
    Load all DataSync profiles defined within configuration.

    :param include_disabled: If true, then disabled profiles will be
       included in the return value; otherwise only currently-enabled
       profiles are returned.
    """
    # Make sure we have a top-level directive.
    keys = get_profile_keys(config, include_disabled=include_disabled)
    if not keys and not ignore_problems:
        raise ConfigurationError(
            "The DataSync configuration does not specify any profiles "
            "to be watched.  Please defined the 'watch' option within "
            "the [rattail.datasync] section of your config file.")

    if include_disabled:
        enabled = get_profile_keys(config, include_disabled=False)

    profiles = {}
    for key in keys:

        if include_disabled:
            try:
                profile = DataSyncProfile(config, key,
                                          load_disabled_consumers=True)
            except Exception as error:
                log.warning("could not create '%s' profile", key, exc_info=True)
                pass
            else:
                profile.enabled = key in enabled
                profiles[key] = profile

        else:
            profile = DataSyncProfile(config, key)
            profile.enabled = True
            profiles[key] = profile

    return profiles



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Email Bounce Handlers
"""

import os
import datetime
import logging
import warnings
from email import message_from_string
from email.utils import parsedate_tz, mktime_tz

from sqlalchemy import orm
from flufl.bounce import all_failures

from rattail.core import Object
from rattail.util import load_object


log = logging.getLogger(__name__)


def get_handler(config, key): # pragma: no cover
    """
    Return a bounce handler instance as specified by configuration.
    """
    warnings.warn("function is deprecated; please use "
                  "`app.get_bounce_handler() instead",
                  DeprecationWarning, stacklevel=2)
    app = config.get_app()
    return app.get_bounce_handler(key)


class Link(Object):
    """
    Simple representation of a link, to be shown in the Tailbone UI to assist
    with processing email bounces.
    """


class BounceHandler(object):
    """
    Default implementation for email bounce handlers.
    """

    def __init__(self, config, config_key):
        self.config = config
        self.config_key = config_key
        self.app = config.get_app()
        self.model = config.get_model()
        self.enum = config.get_enum()

    def get_all_failures(self, msg):
        warnings, failures = all_failures(msg)
        if warnings:
            warnings = ','.join(sorted(
                [e.decode('utf_8') for e in warnings]))
        if failures:
            failures = ','.join(sorted(
                [e.decode('utf_8') for e in failures]))
        return warnings, failures

    @property
    def root_msgdir(self):
        """
        The absolute path of the root folder in which messages are stored.
        """
        return os.path.abspath(self.config.require('rattail.bouncer', 'storage'))

    def msgdir(self, bounce):
        """
        Returns the absolute path of the folder in which the bounce's message
        file resides.  Note that the bounce must already have been persisted to
        the database.  The structure of the path returned is as follows:

        .. code-block:: none

           /{root_msgdir}/{uuid[:2]}/{uuid[2:]}

        * ``{root_msgdir}`` - Value returned by :meth:`root_msgdir()`.
        * ``{uuid[:2]}`` - First two characters of bounce UUID.
        * ``{uuid[2:]}`` - All UUID characters *after* the first two.

        .. note::
           While it is likely that the folder returned by this method already
           exists, this method does not guarantee any such thing.
        """
        return os.path.join(self.root_msgdir, bounce.uuid[:2], bounce.uuid[2:4], bounce.uuid[4:])

    def msgpath(self, bounce):
        return os.path.join(self.msgdir(bounce), 'bounce.eml')

    def handle_bounce_file(self, path, **kwargs):
        """
        Fully "handle" the given email message file, acting on it as
        needed, depending on whether or not it is actually considered
        to be a bounce etc.
        """
        with open(path, 'rt') as f:
            text = f.read()
        msg = message_from_string(text)

        warnings, failures = self.get_all_failures(msg)
        if failures:
            log.info("adding bounce for '%s': %s", path, failures)

            session = self.app.make_session()
            session.continuum_comment = "Received email bounce from {}".format(
                failures)

            bounce = self.make_bounce(msg, failures)
            session.add(bounce)
            session.flush()
            self.store_message_file(bounce, text)
            self.process_bounce(bounce)
            session.commit()

        elif warnings:
            log.warning("found message delivery warning for '%s': %s",
                        path, warnings)
        else:
            log.debug("found message with nothing interesting: %s", path)

    def make_bounce(self, msg, failures=None, **kwargs):
        model = self.model
        if failures is None:
            failed = self.get_failures(msg)
        kwargs.setdefault('config_key', self.config_key)
        kwargs.setdefault('bounced', self.get_bounced_time(msg))
        kwargs.setdefault('bounce_recipient_address', msg['To'])
        kwargs.setdefault('intended_recipient_address', failures)
        return model.EmailBounce(**kwargs)

    def process_bounce(self, bounce):
        pass

    def get_bounced_time(self, msg):
        date = parsedate_tz(msg['Date'])
        if date:
            return datetime.datetime.utcfromtimestamp(
                mktime_tz(date))
        return datetime.datetime.utcnow()

    def store_message_file(self, bounce, msg_body):
        msgdir = self.msgdir(bounce)
        os.makedirs(msgdir)
        path = os.path.join(msgdir, 'bounce.eml')
        with open(path, 'wt') as f:
            f.write(msg_body)
        return path

    def make_link(self, **kwargs):
        return Link(**kwargs)

    def make_links(self, session, recipient):
        model = self.model

        url = self.config.require('tailbone', 'url.customer')
        emails = session.query(model.CustomerEmailAddress).filter_by(address=recipient)
        for email in emails:
            yield self.make_link(type="Rattail Customer",
                                 title=str(email.customer),
                                 url=url.format(uuid=email.parent_uuid))

        url = self.config.require('tailbone', 'url.person')
        emails = session.query(model.PersonEmailAddress).filter_by(address=recipient)
        for email in emails:
            yield self.make_link(type="Rattail Person",
                                 title=str(email.person),
                                 url=url.format(uuid=email.parent_uuid))
